#include <rtai_sched.h>

#include <linux/errno.h>

#include <rtai.h>

#include <rtai_fifos.h>

#include "share_app_tcb.h"

#include "tcb/driver_defs.h"

#include "tcb/messages_formats.h"

#include "tcb/param.h"

#ifdef TTCB
#include "tcb/share_app_ttcb.h"
#endif

#define true			 1
#define false			 0


//when ldegree is not availbale, which
//value should be assumed to estimate
//message delivery delays..in us
#define QOS_UNKWON_LDEGREE_SHIFT    100000	

#define MAX_ITEMS		61      // o maximo possivel - presente no driver de loopback
#define MAX_SEND_TABLE_SIZE	MAX_ITEMS	
#define MAX_RCV_TABLE_SIZE	MAX_ITEMS*NUMBER_OF_MACHINES

// os valores '3' e '100' das linhas seguintes, nao resultam de
// nenhuma derivacao analitica. Como nao e possivel deduzir o tamanho
// maximo das tabelas nesta fase, optamos por valores que nos parecem
// satisfatorios.


#define MAX_WAIT_INFO_TABLE_SIZE MAX_ITEMS*NUMBER_OF_MACHINES*3  	


#define MAX_RCV_PLOAD_TABLE_SIZE 100  	 //tamanho da tabela que guarda informacoes das 
					 //mensagens recebidas no payload no intervalo 	
					 //entre 2 ciclos de leitura.	
// Se durante a execucao do sistema a tabela encher, a TCB simula omissoes 
// no canal de payload para manter a coerencia do sistema.					 


/*
+----------------------------------------+
| tipos das funcoes executaveis pela TCB |
+----------------------------------------+
*/

typedef void (*handtype)(void);	             	      		   //handler de execucao atempada
typedef void (*LFDhandtype)(unsigned int desc);       		   //handler detecao de falhas locais

//handler detecao de falhas distribuidas
typedef void (*DFDhandtype)(unsigned int from, unsigned int seq, unsigned int port); 
//handler da QoS
typedef void (*QoShandtype)(unsigned int port, unsigned long bound, double coverage); 
typedef TCBtimeinterval (*wcetFunctype)(void);    //especificacao de wcet

/*
+--------------------------------------------------------+
| FORMATO DAS ENTRADAS DA TABLELA DE FD HANDLERS REMOTOS |
+--------------------------------------------------------+
*/

typedef struct{
	long seq;
	unsigned char from;
	int port;
}
notification_field_t;

#define REMOTE_NOTIF_LIST_SIZE 100

int remote_notification_indice=0;
notification_field_t pending_notifications[REMOTE_NOTIF_LIST_SIZE];

typedef struct
	{
	  DFDhandtype	  func;          // endereco da funcao a executar
	  TCBtimeinterval wcet; 	 //  
          TCBtimeinterval deadline;	 
          int		  fifo;	
	}
remoteHandInfo;

/*
+-------------------------------------------------------+
| FORMATO DAS ENTRADAS NA TABELA DE EXECUCOES ATEMPADAS |
+-------------------------------------------------------+
*/

typedef struct
	{
	  handtype	  func;          // endereco da funcao a executar
	  int		  fifo;		 // fifo utilizado para comunicar com a aplicacao
	  RT_TASK	  task_id;       // id da task RT que executa handler       
	  TCBtimeinterval wcet;
          TCBtimeinterval delay;
          TCBtimestamp    start_ev;
          TCBtimeinterval max_exec;
          char		  TraceName[21];
#ifdef TTCB
	  int 		   Rindex;
	  tcb_service_in   in;	
#endif          
	}
execInfo;


/*
+-------------------------------------------------------------+
| FORMATO DAS ENTRADAS NA TABELA DE DETECOES DE FALHAS LOCAIS |
+-------------------------------------------------------------+
*/

typedef struct
	{
	  LFDhandtype	  func;            // endereco do handler
	  int		  fifo;            // fifo utilizado para comunicar com a aplicacao
	  char		  faulty;          // flag que activada se der-se a falha
	  RT_TASK	  task_id;	   // id da task RT que executa a funcao
	  TCBtimestamp    start_ev;        // instante inicial do intervalo para a falha
	  TCBtimeinterval spec;            // intervalo temporal para a falha
	  TCBtimeinterval deadline;        // prazo para a execucao
	  char		  TraceName[21];
#ifdef TTCB
	  int 		   Rindex;
	  tcb_service_in   in;	
#endif          
	}
localFDInfo;


/*
+-------------------------------------------------------------+
| TABELAS DO SERVICO DE DETECAO DE FALHAS DISTRIBUIDAS	      |
+-------------------------------------------------------------+
*/

/*
Nesta tabela e guardada a informacao relativa as mensagens recebidas no
canal de payload, e cuja informacao de falha ainda nao foi recebida pela
TCB (no canal de controlo).
*/
typedef struct
	{
		unsigned long long clean_up_after; // instante apos o qual o colector de lixo pode "limpar" a entrada
		unsigned long long time_of_receipt;// instante em que a mensagem foi recebida 
		unsigned char from;                // id da TCB da maquina emissora
		unsigned long seq;		   // numero de sequencia 
		unsigned char is_free;              
	}
rcv_payload_info;	

/*
Tabela que guarda os pedidos de detecao de falhas distribuidas solicitados a TCB
ate que a tarefa de envio seja activada e consequentemente a informacao disseminada
no canal de controlo.
*/

typedef	struct
	{
		unsigned long long time_of_generation;  // instante que assinala o inicio da medicao
		unsigned long spec;			// tempo ate falhar (em us)
		unsigned long seq;			// numero de sequencia da mensagem
		unsigned char to;			// mascara que especifica os destinatarios da mensagem
		unsigned int  destination_port;         // porto destino (porto UDP) ao qual a mensagem e destinada
		unsigned char remote_tfd;		// remote detection?
		unsigned char is_free;
	}
snd_info;

/*
Tabela que guarda os pedidos de detecao de falhas distribuidas recebidos no
canal de controlo. Alguns campos sao especificados aquando da recepcao do pedido
*/

typedef	struct
	{
		unsigned long long time_of_failure; // instante da falha (estimado no relogio do receptor)
		unsigned long seq;                  // numero de sequencia da mensagem
		unsigned char from;   		    // id da TCB da maquina emissora	
		unsigned char to;		    // mascara que especifica os destinatarios da mensagem
		unsigned char is_free;
		unsigned int  destination_port;     // porto destino (porto UDP) ao qual a mensagem e destinada
		char fail;			    // falha temporal? 	
		unsigned long lateness_degree;      // grau de atraso  
		char lateness_available;            // o grau de atraso esta disponivel?
		char remote_tfd;		    // deteccao remota?
	}
rcv_info;	

typedef struct
	{
		unsigned long long failure_at_time;  // instante da falha 
		unsigned char is_free;
		int           send_fifo;             //  envia a informacao para a 
						     //  aplicacao que enviou a mensagem 
						     //  (-1 significa que nao e a maquina emissora)	
		unsigned long seq;
		int           owner; 
//		char 	      exec;
		unsigned char n_dests;               //  numero de destinatarios da mensagem
		unsigned char from;		
		unsigned char existsTask;	
		RT_TASK       rt_task_id;	     // id da tarefa TR que faz a detecao atempada da 
						     // enventual falha temporal (so difinido na maquina emissora da mensagem)	
		DFDhandtype function;			     
		unsigned char to;
		unsigned char any_fail;		     // ja foi recebida alguma notificacao positiva de falha 
		unsigned int  destination_port;
		fail_info     failure_info[NUMBER_OF_MACHINES];  
		unsigned int  spec;
	}
wait_info;	


/*
+-------------------------------------------------------------+
| MECANISMO DE TRACING - EM CONSTRUCAO			      |
+-------------------------------------------------------------+
*/


typedef struct
	{
		unsigned long seq;
		unsigned long long time;
		unsigned int  evt;
		unsigned long arg;
		char	      strarg[20]; 	
	}
trace_msg;


/*
+-------------------------------------------------------------+
| RELOGIO GLOBAL					      |
+-------------------------------------------------------------+
*/

typedef struct{
	RTIME	vclock;		//valor do relogio virtual
	double	error;
	RTIME   time_of_sync;
}global_clock_info;



/*
+-------------------------------------------------------------+
| QoS ADAPTATION 					      |
+-------------------------------------------------------------+
*/

#define N_SAMPLES	100

typedef struct{
	QoShandtype	   func;          // endereco da funcao a executar	
	unsigned long	   dev;
	double last_value;	
	double		   coverage;
	unsigned long	   bound;
	char 		   mode; 	// 0 - awareness, 1 - stability
	//samples
	double samples[N_SAMPLES];
	int head, tail;
	double average;
	double variance;
	unsigned int fifo;
	char used;	
	char notification;
}QoS_Adaptation_t;




