/*	 
	(T)TCB kernel

--[TTCB 1.1]----------------------------------------------------

    (Trusted) Timely Computing Base 1.1
    (C) 1999-2003, Navigators Group 
    		   http://www.navigators.di.fc.ul.pt
    
    ------------------------------------------------------------   
    
    Pedro Martins <pmartins@di.fc.ul.pt>
    	core TCB kernel and services 	
    
    Miguel Correia <mpc@di.fc.ul.pt>
    	TTCB extensions and services
    
    Nuno Miguel Neves <nnveves@di.fc.ul.pt>
    	TTCB extensions and services
     

-----------------------------------------------------------------	
*/

/*
 * DEFINES **************************************************************
 *
 */

//uncomment the following line for a release version
//#define RELEASE

//#define DEBUG_VERSION

#if !defined (RELEASE) && !defined(BENCH)
#define DEBUGNET
#endif

//#define GLOBAL_CLOCK_DEBUG

/* ********************************************************************* */

#include <linux/config.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#	define MODVERSIONS 
#endif

#ifdef MODVERSIONS
#	include <linux/modversions.h>
#endif


#ifndef MODULE
#define MODULE
#endif


#include <linux/module.h>
#include <asm/module.h>
#include <asm/io.h>

#include "tcb/defs.h"
#include "tcb/trace_defs.h"
#include <rtai_fifos.h>
#include <rtai_sched.h>
#include <rtai.h>
#include "tcb/TCBnetdevice.h"
#include "tcb/version.h"

#ifdef TTCB
#include "../crypto/rsa/mpi.h"
#include "../crypto/rsa/rsa.h"
#include "tcb/ttcb_module.h" 

RSA_secret_key SecretKey;
RSA_public_key PublicKey;

void hash_md5(unsigned char *in, int inlen, unsigned char *key, int keylen, unsigned char *out);
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR ("Pedro Martins, Miguel Correia, Nuno Miguel Neves");
MODULE_DESCRIPTION ("(Trusted) Timely Computing Base for RTAI");  

#ifdef DISTRIBUTED
void openReceive(void);
void closeReceive(void);
#endif

#define SYSTEM_TASK_PRIORITY	1
#define USER_TASK_PRIORITY	2

/*
+-------------------------------+
|  PRIMITIVAS DE SINCRONIZACAO  |*********************************************
+-------------------------------+ 
*/

// NOTA: nao podem ser anichadas (nested)

int curr_priority;


void TCBenter_in_critical_section(void){
	//e o linux?		
	if (rt_whoami()->priority > USER_TASK_PRIORITY)
	{
		cli();
	}
	curr_priority = rt_whoami()->priority;
	rt_change_prio(rt_whoami(),SYSTEM_TASK_PRIORITY-1);		
}

void TCBleave_critical_section(void){
	if (rt_whoami()->priority > USER_TASK_PRIORITY)
	{
		sti();
	}
	rt_change_prio(rt_whoami(),curr_priority);		
}

/*
+----------------------+
|  MENSAGENS DE DEBUG  |******************************************************
+----------------------+ 
*/

/*
	-----------
	DEBUG level
	-----------
		1 - view events (joining and leaving of nodes) 
		2 - service invocations
		4 - network packets
		16 - TFD invocations/replies
		8 -
		
*/

#define VIEW_LEVEL 1
#define SERVICE_LEVEL 2
#define NETWORK_LEVEL 4
#define TFD_LEVEL 16
#ifdef TTCB
#define DECISAO_LEVEL 8
#define SECURITY_LEVEL 32
#define INTERNAL_LEVEL 64
#endif

//change this def. to specify a new debug level
//#define DEBUG_LEVEL (TFD_LEVEL) 	

#define DEBUG_LEVEL 0

#if DEBUG_LEVEL!=0
#define DEBUG_MSG(level, fmt, args...) if (DEBUG_LEVEL & level) rt_printk(fmt,##args) 
#else 
#define DEBUG_MSG(level, fmt, args...)
#endif

/*
+--------------------+
|  VARIAVEIS GLOBAIS |********************************************************
+--------------------+ 
*/

tcb_kernel_info_t	kern_info;
 


#ifdef GLOBAL_CLOCK_DEBUG
	char gc_debug_code=1;
#endif

#ifdef BENCH
	unsigned long bench_max_exec_read=0,bench_max_exec_send=0;
	unsigned long max_roundtrip=0;
	unsigned long long next_release_of_send;
	unsigned long long next_release_of_read;
	unsigned long max_release_time_send=0;
	unsigned long max_release_time_read=0;
	unsigned long max_release_delay_of_handler_DFD=0; 
	unsigned long max_release_delay_of_handler_LFD=0; 
	unsigned long max_global_clock_drift=0; 
#endif

unsigned long   seq_numbers=1;    // 23 bits

char 		usedFifos
		[MAX_TCB_CONNECTIONS];   //tabela que contem a informacao sobre 				         //a utilizacao dos fifos(e.g.conexoes)				         
	
#ifdef LOCAL_MEASURE
RTIME		localMeasures
                [MAX_LOCAL_MEASURES];    //tabela de medicoes locais                 
#endif //LOCAL_MEASURE

#ifdef LOCAL_TIMELY
execInfo 	executions
                [MAX_EXEC];              //tabela de execucoes atempadas
#endif //LOCAL_TIMELY

#ifdef LOCAL_FAULT
localFDInfo 	localFD
                [MAX_LOCAL_FD];          //tabela de deteccoes de falhas locais
#endif //LOCAL_FAULT

#ifdef DISTRIBUTED
global_clock_info	gclock_info[NUMBER_OF_MACHINES];
RTIME			vclock_at_sync;  
RTIME			time_of_sync;
double			virtual_drift;
RTIME			apply_drift_until;
#endif

/*
+------------+
| NETWORKING |
+------------+       
*/

unsigned char   tcb_machine_id;
unsigned long	trace_msg_number=0;

//VER COM O MIGUEL/NUNO TTCBcrashedLocal!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!|

char 		system_view[NUMBER_OF_MACHINES]; // 0-NOT_PRESENT, 1-IN_INIT, 2-NORMAL

#ifdef DISTRIBUTED

char 		state_of_operation; // init, normal ...
unsigned char   contention_rounds;
unsigned char   number_of_active_machines=0;

#define  MAX_MTU	1600

unsigned char   net_msg_memory[MAX_MTU]; // network message memory
RT_TASK		snd_task_id,rcv_task_id;
msg_info_type	last_rcv_info[NUMBER_OF_MACHINES];    // informacao das ultimas mensagens recebidas de cada no
char            wait_round; 

unsigned long   read_loops=0;
unsigned char   my_mask;		
unsigned int	netMTU;
RTIME		start_of_system_tasks;

int		 max_fields,max_items;
int 		 n_sends_in_this_slot;

rcv_info	 receive_info[MAX_RCV_TABLE_SIZE];

unsigned char	invalid_msg_display=0;
int		od_counter=0;
unsigned long long send_time;

#endif

#ifdef DIST_FAULT
/*
+-------------------------------+
| DISTRIBUTED FAILURE DETECTION |
+-------------------------------+       
*/

rcv_payload_info receive_payload_info[MAX_RCV_PLOAD_TABLE_SIZE];
int 		 number_of_detections[NUMBER_OF_MACHINES];
snd_info	 send_info[MAX_SEND_TABLE_SIZE];
wait_info	 wait_infos[MAX_WAIT_INFO_TABLE_SIZE];

int   	         port_to_fifo[65536];	//arranjar uma solucao + elegante!

#endif // DIST_FAULT



#ifdef DISTRIBUTED // The global clock only makes sense ina distributed environment
/*
+----------------------+
|  RELOGIO GLOBAL      |******************************************************
+----------------------+ 
*/

// Retorna o valor do relogio virtual no instante 'time' dado pelo relogio fisico
// NOTA: time <= rt_get_time_ns()
RTIME get_vclock_at(RTIME time){	
	long long elapsed;	

	if (virtual_drift == 0){
		elapsed = time - time_of_sync;	
		return vclock_at_sync + elapsed;
	}
	
	if (time > apply_drift_until){
		elapsed = time - apply_drift_until;	
		vclock_at_sync = vclock_at_sync + ((double)(apply_drift_until-time_of_sync)*virtual_drift) + (apply_drift_until - time_of_sync); 	
		virtual_drift = 0;		
		time_of_sync = apply_drift_until;
		return vclock_at_sync + elapsed;	
	}	
	elapsed = time - time_of_sync;	

	return (vclock_at_sync + elapsed + ((double)elapsed * virtual_drift));
}

// Retorna o valor do relogio virtual no instante actual
RTIME get_vclock(void){
	return get_vclock_at(rt_get_time_ns());
}

void init_vclock(void){
	int a;
	RTIME t;

	for (a=0;a<NUMBER_OF_MACHINES;a++){
		if (gclock_info[a].vclock!=0){
			break;
		}
	}

	if (a == NUMBER_OF_MACHINES){
		virtual_drift = 0;	
		t = rt_get_time_ns();
		vclock_at_sync = t;
		time_of_sync = t;
	}
}

// actualiza vclock
void update_vclock(unsigned char from, RTIME from_vclock, unsigned long rndtrip, RTIME rcv_time){
	int a;
	RTIME curr_error = 0;

	if (from != tcb_machine_id){
		if (rndtrip!=0){		
			curr_error = gclock_info[from].error + ((double)(rcv_time-gclock_info[from].time_of_sync)*2*CLOCK_DRIFT_RATE);
			if (curr_error > ((rndtrip >> 1) * 1000) )
			{
				gclock_info[from].vclock = from_vclock + ((rndtrip >> 1) * 1000);	
				gclock_info[from].error = ((rndtrip >> 1) * 1000);
				gclock_info[from].time_of_sync = rcv_time;
			}
			else{
				return ;
			}
		}
		else {
			gclock_info[from].vclock = from_vclock + ((MAXIMUM_ROUNDTRIP >> 1) * 1000);	
			gclock_info[from].error = ((MAXIMUM_ROUNDTRIP >> 1) * 1000);
			gclock_info[from].time_of_sync = rcv_time;
			return ;
		}
	}
	
	if (state_of_operation == NORMAL_STATE){

	for (a=0;a<NUMBER_OF_MACHINES;a++){
		if ((gclock_info[a].vclock!=0)||(a == tcb_machine_id)){
			if (from == a){
				vclock_at_sync = get_vclock_at(gclock_info[a].time_of_sync);
				time_of_sync = gclock_info[a].time_of_sync;
				if (from == tcb_machine_id)					
					virtual_drift = 0;
				else
				{	
					virtual_drift = (double)(gclock_info[a].vclock - vclock_at_sync) / SEND_PERIOD;
					apply_drift_until = time_of_sync + SEND_PERIOD;
#ifdef BENCH
	if (curr_error > max_global_clock_drift)
		max_global_clock_drift = curr_error;
#endif		
				}
			}
			break;
		}
	
	}
	}

	if (state_of_operation != NORMAL_STATE){

	for (a=0;a<NUMBER_OF_MACHINES;a++){
		if (gclock_info[a].vclock!=0){
			if (from == a){
				vclock_at_sync = gclock_info[a].vclock;
				time_of_sync = gclock_info[a].time_of_sync;;
				virtual_drift = 0;
			}
			break;
		}
	
	}
	}	
}
#endif //DISTRIBUTED 


#ifdef DISTRIBUTED
/*
+----------------------+
|  FUNCOES AUXILIARES  |******************************************************
+----------------------+ 
*/

/*
+-----------------------------------------+
| Retorna o numero de maquinas do sistema |
| que estao no estado normal.		  |
+-----------------------------------------+
*/

inline unsigned char get_number_of_normal_state_machines(void)
{
	unsigned char res=0;
	int a;
	
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	{
		if (system_view[a]==NORMAL)
		  res++;
	}
	
	return res;
}

#endif //DISTRIBUTED

#ifdef TTCB
/*
+--------------------+
|  SERVICOS DA TTCB  |**********************************************************
+--------------------+ 
		

+----------------------------------------+
| enviar chave pblica da TTCB           |
+----------------------------------------+
*/

void TTCBGivePublicKey(int fifo)
{
  int sizeE;
  char sendBuffer[200]="";
  int put_return;

  DEBUG_MSG(SECURITY_LEVEL,"Going to give my PublicKey!\n");
  
  
  
  /*   Sizes:
       PublicKey : 8
       N is always 1024, unless we change the key length
       So the rest is the exponent
  */
  sizeE=mpi_get_nbits(PublicKey.e);
  /* sendBuffe=(char *)vmalloc(1024+sizeE);*/
  memcpy(sendBuffer, PublicKey.n, 24);
  memcpy(sendBuffer+24, PublicKey.n->d,128);

  memcpy(sendBuffer+24+128, PublicKey.e, 24);
  memcpy(sendBuffer+128+48,PublicKey.e->d,(sizeE+7)/8);

  
  put_return=rtf_put(fifo, sendBuffer, 128 + 48 + (sizeE+7)/8 );

  DEBUG_MSG(SECURITY_LEVEL,"Sent Public Key!");

  return;
}
#endif

/*
+-------------------+
|  SERVICOS DA TCB  |**********************************************************
+-------------------+ 
		
+-----------------------------------------------+
| enviar estampilha temporal global a aplicacao |
+-----------------------------------------------+
*/
#ifdef DISTRIBUTED //There is no global_timestamp with no distribution
#ifdef TTCB
void TCBget_global_timestamp(int fifo, tcb_service_in *  in)
#else
void TCBget_global_timestamp(int fifo)
#endif
{
	RTIME t;
#ifdef TTCB
        tcb_service_out tsout;
	unsigned char Rindex;
#endif
#ifdef GLOBAL_CLOCK_DEBUG
	unsigned long irqstate;			
	
	irqstate = rt_global_save_flags_and_cli();		   	
	outb(gc_debug_code,0x378);
	gc_debug_code++;
	if (gc_debug_code==3)
	 gc_debug_code=1;
#endif

	TCBenter_in_critical_section();

	t = get_vclock_at(rt_get_time_ns());

	TCBleave_critical_section();

#ifdef GLOBAL_CLOCK_DEBUG
	rt_global_restore_flags(irqstate);			   		
#endif
#ifdef TTCB

	*((RTIME*)(&tsout.buffer))=t;
	tsout.header_out.error=0;
	Rindex=EIDtoINDEX(in->header_in.eid);
	process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else
	rtf_put(fifo,&t,sizeof(RTIME));		
#endif //TTCB
}

/*
+-----------------------------------------------------+
| enviar estampilha temporal local/global a aplicacao |
+-----------------------------------------------------+
*/

#ifdef TTCB
void TCBget_global_local_timestamp(int fifo, tcb_service_in *  in)
#else
void TCBget_global_local_timestamp(int fifo)
#endif
{
	local_global_ts_t t;
#ifdef TTCB
        tcb_service_out tsout;
	unsigned char Rindex;
#endif
	TCBenter_in_critical_section();
	t.ts = rt_get_time_ns();
	t.gts = get_vclock_at(t.ts);
	TCBleave_critical_section();

#ifdef TTCB

	*((local_global_ts_t*)(&tsout.buffer))=t;
	tsout.header_out.error=0;
	Rindex=EIDtoINDEX(in->header_in.eid);
	process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else
	rtf_put(fifo,&t,sizeof(local_global_ts_t));		
#endif
}

#endif //DISTRIBUTED

/*
+-----------------------------------------+
| enviar informacao do kernel a aplicacao |
+-----------------------------------------+
*/
#ifdef TTCB
void TCBget_kernel_info(int fifo, tcb_service_in *  in)
#else
void TCBget_kernel_info(int fifo)
#endif
{
#ifdef TTCB
        tcb_service_out tsout;
	unsigned char Rindex;
#endif	
#ifdef TTCB

	memcpy((&tsout.buffer), &kern_info, sizeof(tcb_kernel_info_t));
	tsout.header_out.error=0;
	Rindex=EIDtoINDEX(in->header_in.eid);
	process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else
	rtf_put(fifo,&kern_info,sizeof(tcb_kernel_info_t));		
#endif

}


/*
+----------------------------------------+
| enviar estampilha temporal a aplicacao |
+----------------------------------------+
*/
#ifdef TTCB
void TCBget_timestamp(int fifo, tcb_service_in *  in)
#else
void TCBget_timestamp(int fifo)
#endif
{
	RTIME t;
#ifdef TTCB
        tcb_service_out tsout;
	unsigned char Rindex;
#endif

	t=rt_get_time_ns();

#ifdef TTCB

	*((RTIME*)(&tsout.buffer))=t;
	tsout.header_out.error=0;
	Rindex=EIDtoINDEX(in->header_in.eid);
	process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else
	rtf_put(fifo,&t,sizeof(RTIME));		
#endif
}

/*
+----------------------------------------+
| obter o identificador da TCB (nodo)	 |
+----------------------------------------+
*/

unsigned char get_tcb_id(void)
{
#ifdef DISTRIBUTED
	u64 macAddr=getMacAddr();        // obter endereco ethernet 
	  	  		  	 // da placa de rede do canal
					 // sincrono	   
	switch (macAddr) {
		#include "tcb/rt_net_addrs.h"
	}
#endif
	return 0;
}

/*
+------------------------------+
| enviar ID da TCB a aplicacao |
+------------------------------+
*/

#ifdef TTCB
void TCBget_id(int fifo, tcb_service_in *  in)
#else
void TCBget_id(int fifo)
#endif
{	
#ifdef TTCB
        tcb_service_out tsout;
	unsigned char Rindex;
#endif		
	int t=(int)get_tcb_id();
#ifdef TTCB

	*((int*)(&tsout.buffer))=t;
	tsout.header_out.error=0;
	Rindex=EIDtoINDEX(in->header_in.eid);
	process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else
	rtf_put(fifo,&t,sizeof(t));		
#endif
}



#ifdef DIST_FAULT
/*
+--------------------------------------------+
| handler de deteccao de falhas distribuidas |
+--------------------------------------------+
*/

void failTask(int par)   
{
#ifdef BENCH
	RTIME tof=wait_infos[par].failure_at_time;
	RTIME t;
	long long diff;

	t=rt_get_time_ns();
	diff=t-tof;
	if (diff>max_release_delay_of_handler_DFD)
		max_release_delay_of_handler_DFD=diff;
#endif	//BENCH

	DEBUG_MSG(TFD_LEVEL,"Executing TFD handlder: any_fail <%u> seq <%u>\n",wait_infos[par].any_fail,wait_infos[par].seq);

	wait_infos[par].function(wait_infos[par].from,wait_infos[par].seq,wait_infos[par].destination_port);

//avoid preemption here
	rt_change_prio(rt_whoami(),SYSTEM_TASK_PRIORITY-1);	
			
	if (wait_infos[par].any_fail==2){
        	wait_infos[par].is_free=1;  
        }
        
        wait_infos[par].existsTask = 0;
	rt_task_delete(rt_whoami());	    
}
#endif // DIST_FAULT


/*
+----------------------------------------+
| iniciar deteccao de falha distribuida	 |
+----------------------------------------+
*/

#ifdef DIST_FAULT

#ifdef TTCB
void TCBstart_distributedFD(int fifo, tcb_service_in *  in)
#else //TTCB
void TCBstart_distributedFD(int fifo,distributedFDPars* pars)  
#endif //TTCB
{
	long t;
	int a,c,ret,sndPt=0;
	char to_many=0;
	unsigned long irqstate=0;			
	char to_array[NUMBER_OF_MACHINES];		
	unsigned char mask=1;	
	unsigned char dest;
	unsigned char number_of_dest=0;
	DFDhandtype     funcptr;
	wcetFunctype wcetptr;
#ifdef TTCB
	distributedFDPars* pars;
	int Rindex;
	tcb_service_out tsout;
	Rindex=EIDtoINDEX(in->header_in.eid);
	pars = (distributedFDPars*)in->buffer;
	tsout.header_out.error=0;
#endif //TTCB
	dest=pars->to;

	TCBenter_in_critical_section();

	if (n_sends_in_this_slot==max_items)
	{	
		TCBleave_critical_section();
		t=BURSTY_SENDS;
#ifdef TTCB
		*((long*)(&tsout.buffer))=t;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else	 //TTCB
	        rtf_put(fifo,&t,sizeof(unsigned long));		
#endif //TTCB
	        return;				
	}

	TCBleave_critical_section();
	
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	{
		if (mask&dest)
		{
		  to_array[a]=1;
		  number_of_dest++;
		}  
		   else
		    to_array[a]=0;
		mask<<=1;
	}
	
	if (strcmp(pars->modName,"")!=0)
	{
	
	//obter endereco da funcao	

	funcptr=(DFDhandtype)inter_module_get_request(pars->funcName,pars->modName);
	if (funcptr != 0){
		inter_module_put(pars->funcName);
	}
	else
	{
		t=FUNC_NOT_FOUND;
#ifdef TTCB
		*((long*)(&tsout.buffer))=t;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else	//TTCB
		rtf_put(fifo,&t,sizeof(t));		
#endif	//TTCB
		return;
	}
		
	wcetptr=(wcetFunctype)inter_module_get_request(pars->wcetName,pars->modName);
		
	if (wcetptr != 0){
		inter_module_put(pars->wcetName);
	}
	else
	{
		t=WCET_FUNC_NOT_FOUND;
#ifdef TTCB
		*((long*)(&tsout.buffer))=t;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else	//TTCB
		rtf_put(fifo,&t,sizeof(t));		
#endif	//TTCB
		return;
	}
	}
	 else
	  funcptr=0;

	//sincronizar o acesso com a tarefa de leitura
	//introduz uma latencia controlada na execucao da tarefa de leitura
	TCBenter_in_critical_section();
	
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	  if (to_array[a]==1)
	   if (number_of_detections[a]+1>max_items)	
	   {
	    to_many=1;
	    break;
	   } 
	
	TCBleave_critical_section();	
	
	if (to_many)
	{
	 t=TOO_MANY_DISTRIBUTED_FD;
#ifdef TTCB
		*((long*)(&tsout.buffer))=t;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else //TTCB
	 rtf_put(fifo,&t,sizeof(unsigned long));		
#endif //TTCB
	 return;
	} 
		
	t=0;

	TCBenter_in_critical_section();
			
	n_sends_in_this_slot++;
			
	//find a free entry on the send_info table		  
	for (a=0;a<MAX_SEND_TABLE_SIZE;a++)
	{
		if (send_info[a].is_free)
		 {
			t=1;
			send_info[a].is_free=0;
			send_info[a].time_of_generation=pars->start_ev;
			send_info[a].spec=pars->spec;
			send_info[a].destination_port=pars->destination_port;
			send_info[a].seq=seq_numbers;
			send_info[a].to=pars->to;
			sndPt=a;
			break;	 		
		 }
	}
	
	if (t!=0)
	{
	t=0;
	// iniciar a estrutura do WaitInfo
	for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++)
	{
		if (wait_infos[a].is_free)
		 {
			t=1;
			if (funcptr!=0)
			{
			
		 	 wait_infos[a].function=funcptr;				 
			 irqstate = rt_global_save_flags_and_cli();		   	 
			 wait_infos[a].failure_at_time=(pars->spec*1000)+pars->start_ev;
				
			if (wait_infos[a].failure_at_time<rt_get_time_ns())
			{
				send_info[sndPt].is_free=1;	       	          	
				rt_global_restore_flags(irqstate);
				TCBleave_critical_section();		
				t=ALREADY_FAILED;
#ifdef TTCB
				*((long*)(&tsout.buffer))=t;
				process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else		
	       	          	rtf_put(fifo,&t,sizeof(t));	       	          		       	          	
#endif
	       	          	return;
				
			}
	
			ret=rt_task_init (&(wait_infos[a].rt_task_id),failTask,a,4000,USER_TASK_PRIORITY,1,0);
			if (ret!=0)
	 		 {
	 		  send_info[sndPt].is_free=1;			  	
			  rt_global_restore_flags(irqstate);
	       	          TCBleave_critical_section();
	    		  t=NO_AVAIL_EXEC;
#ifdef TTCB
 			  *((long*)(&tsout.buffer))=t;
			  process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else
	       	          rtf_put(fifo,&t,sizeof(t));	       	          	       	          
#endif
	       	          return;
		         }
	                 else {
			   wait_infos[a].existsTask=1;		          
			  }  
		       } //funcptr!=0    
	       		else
	          	 wait_infos[a].existsTask=0;		          

			wait_infos[a].is_free=0;
			wait_infos[a].owner=fifo;
			wait_infos[a].send_fifo=fifo+INC_TO_WAIT_FIFO;
			wait_infos[a].seq=seq_numbers;
			wait_infos[a].to=pars->to;
			wait_infos[a].any_fail=0;
		 	wait_infos[a].n_dests=number_of_dest;
		 	wait_infos[a].from=tcb_machine_id;
		 	wait_infos[a].destination_port=pars->destination_port;   
		 	for (c=0;c<NUMBER_OF_MACHINES;c++)
		 	 wait_infos[a].failure_info[c].fail=2;
	 		for (c=0;c<NUMBER_OF_MACHINES;c++)
			{
			  if (to_array[c]){
			   number_of_detections[c]++;
			  }	
			}
		
	 	        rt_global_restore_flags(irqstate);
			break;	 		
		 }
	}
	}
	
	TCBleave_critical_section();	
	
	if (t==0)  //should never happens if the system is correct
	 while (1){
	 	rt_printk("TTCB PANIC: no free send_info or wait_info\n");
	 }
	
	t=seq_numbers;
#ifdef TTCB
        *((long*)(&tsout.buffer))=t;
        process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else
	rtf_put(fifo,&t,sizeof(unsigned long));		
#endif
	
	seq_numbers++;
	
	if (seq_numbers==8388608)  //23 bits bound
	 seq_numbers=1;
	
}
#endif //DIST_FAULT


/*
+-----------------------------------------------------------+
| inicia medicao local e devolve um descriptor a aplicacao. |
| retorna NO_AVAIL_LOCAL_MEASURE se nao existir descriptor  |
| disponivel.				       	            |
+-----------------------------------------------------------+
*/
#ifdef LOCAL_MEASURE

#ifdef TTCB
void TCBstart_local_measure(int fifo,tcb_service_in *in)  //PARA SAIR
#else //TTCB
void TCBstart_local_measure(int fifo,RTIME *initTS)  //PARA SAIR
#endif //TTCB
{
	int a,desc=NO_AVAIL_LOCAL_MEASURE;
#ifdef TTCB
	RTIME *initTS;
	int Rindex;
	tcb_service_out tsout;
	Rindex=EIDtoINDEX(in->header_in.eid);
	initTS = (RTIME *)in->buffer;
	tsout.header_out.error=0;
#endif //TTCB
	
	for (a=0;((a<MAX_LOCAL_MEASURES)&&(desc==NO_AVAIL_LOCAL_MEASURE));a++)
	{
		if (localMeasures[a]==0)
		{
			desc=a;
			localMeasures[a]=*initTS;
		}
	}
	
#ifdef TTCB
	*((int*)(&tsout.buffer))=desc;
	process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else //TTCB
	rtf_put(fifo,&desc,sizeof(int));
#endif //TTCB
}


/*
+----------------------------------------------------------------+
| termina medicao local; liberta descriptor; devolve a aplicacao |
| a medida da duracao e a estampilha final                       |  
| retorna INVALID_LOCALM_DESC se o descriptor for invalido       | 
+----------------------------------------------------------------+
*/

#ifdef TTCB
void TCBend_local_measure(int fifo,tcb_service_in *in)  //PARA SAIR
#else //TTCB
void TCBend_local_measure(int fifo,int *descriptor)  //PARA SAIR
#endif //TTCB
{
	endLocalMResult res;
	RTIME actTime;
#ifdef TTCB
	int *descriptor;
	int Rindex;
	tcb_service_out tsout;
	Rindex=EIDtoINDEX(in->header_in.eid);
	descriptor = (int *)in->buffer;
	tsout.header_out.error=0;
#endif	//TTCB
	/* descriptor invalido ??? */
	if ( (*descriptor<0) || (*descriptor>=MAX_LOCAL_MEASURES) || (localMeasures[*descriptor]==0) ) 
	{
		res.stat=INVALID_LOCALM_DESC;
#ifdef TTCB
		memcpy((&tsout.buffer),&res, sizeof(res));
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else	//TTCB
		rtf_put(fifo,&res,sizeof(res));
#endif	//TTCB
		return;
	}
	
	actTime=rt_get_time_ns();
	
	// obter numero de nanosegundos desde o inicio da medicao
	res.tm=actTime-localMeasures[*descriptor];
	
	res.ts=actTime;
	res.stat=OK;

#ifdef TTCB
	memcpy((&tsout.buffer),&res, sizeof(res));
	process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else //TTCB	
	rtf_put(fifo,&res,sizeof(res));
#endif //TTCB
	//libertar descriptor
	localMeasures[*descriptor]=0;	
}

#endif //LOCAL_MEASURE

/*
+--------------------------------------------------------------------------+
|                           EXECUCAO ATEMPADA                              |**
+--------------------------------------------------------------------------+**
   ***************************************************************************
*/


#ifdef LOCAL_TIMELY 
/*
+----------------------------------------------------------------+
| verifica se e executavel; recalcula delay e adormece newdelay; |
| executa a funcao e devolve a aplicacao o instante final        |
+----------------------------------------------------------------+
*/

void execTask(int par)
{
	RTIME now;
	RTIME nanos_past_start_ev;
	RTIME newdelay;
	long long res;
	RTIME end_ev,a,b;
	int desc=par;	
	unsigned long irqstate;
#ifdef TTCB
	int Rindex = executions[desc].Rindex;
	tcb_service_out tsout;
	tcb_service_in * in = &(executions[desc]. in);
	tsout.header_out.error = 0;
#endif
	
	irqstate = rt_global_save_flags_and_cli();		   
		
	//instante actual
	now=rt_get_time_ns();
	
	// nanosegundos desde o instante inicial
	nanos_past_start_ev=now-executions[desc].start_ev;
	
	//recalcular o delay
	 if (executions[desc].delay < nanos_past_start_ev )
	   newdelay=0;
	  else
	   newdelay=executions[desc].delay-(nanos_past_start_ev);

	a=now+newdelay+executions[desc].wcet;
	b=executions[desc].start_ev + executions[desc].max_exec;


//MUDAR PARA MSG_DEBUG!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#ifdef DEBUG
	rt_printk("Exec:\nStart Ev =");
	printTimeStamp(executions[desc].start_ev);
	rt_printk("\nDelay =");
	printTimeStamp(executions[desc].delay);
	rt_printk("\nWcet =");
	printTimeStamp(executions[desc].wcet);
	rt_printk("\nMax Exec =");
	printTimeStamp(executions[desc].max_exec);
	rt_printk("\nNewDelay =");
	printTimeStamp(newdelay);
	rt_printk("\nNow =");
	printTimeStamp(now);
	rt_printk("\n");
#endif
	
	// verificar se e possivel executar atempadamente
	if ( a > b )
	{
	  rt_global_restore_flags(irqstate);			   
	  //nao e possivel executar atempadamente
	  res=CANT_RUN_TIMELY;
#ifdef TTCB
	  *(long long*)(&tsout.buffer)=res;
	  process_output(executions[desc].fifo, (out_hdr_t *)&tsout, sizeof(tcb_service_out), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else					
	  rtf_put(executions[desc].fifo,&res,sizeof(res));		
#endif
	  executions[desc].func=0;
	  rt_task_delete(rt_whoami());	    
	}
	 else
	  {
	     rt_task_make_periodic(rt_whoami(),nano2count(executions[desc].delay+executions[desc].start_ev),0);	
	     rt_global_restore_flags(irqstate);			   	 			     	
	     rt_task_wait_period();                                                                          
	     	     
	   executions[desc].func(); 	 // executa a funcao	
	   end_ev=rt_get_time_ns();      //obtem o instante final
#ifdef TTCB
	  *((RTIME*)&(tsout.buffer))=end_ev;										
	  process_output(executions[desc].fifo, (out_hdr_t *)&tsout, sizeof(tcb_service_out), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else	   	//TTCB
	   rtf_put(executions[desc].fifo,&end_ev,sizeof(end_ev)); //envia a aplicacao				
#endif	//TTCB
 	   executions[desc].func=0; //liberta entrada na tabela
	  }
	
	rt_task_delete(rt_whoami());	    
}



/*
+---------------------------------+
| inicia execucao atempadada      | 
+---------------------------------+
| retorna a aplicacao:            |
|          timestamp              |
|	   FUNC_NOT_FOUND	  |	
|	   WCET_FUNC_NOT_FOUND	  |			
|	   NO_AVAIL_EXEC          |
+---------------------------------+
*/
#ifdef TTCB
void TCBexec(int fifo, tcb_service_in *in)  
#else	
void TCBexec(int fifo, execPars* pars)
#endif
{	
	handtype     funcptr;
	wcetFunctype wcetptr;
	long long res;
	int a,desc=NO_AVAIL_EXEC;
	int ret;
#ifdef TTCB
	execPars *pars;
	int Rindex;
	tcb_service_out tsout;
	Rindex=EIDtoINDEX(in->header_in.eid);
	pars = (execPars *) in->buffer;
	tsout.header_out.error=0;
#endif	//TTCB	
	
	// procurar descriptor livre
	for (a=0;((a<MAX_EXEC)&&(desc==NO_AVAIL_EXEC));a++)
	{
		if (executions[a].func==0)
			desc=a;
	}
	
	if (desc == NO_AVAIL_EXEC)
	{
		res=NO_AVAIL_EXEC;
#ifdef TTCB
		*(TCBtimestamp*)(&tsout.buffer)=res;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else //TTCB			
		rtf_put(fifo,&res,sizeof(res));		
#endif	//TTCB
		return;
	}
	
	//obter endereco da funcao	
	funcptr=(handtype)inter_module_get_request(pars->funcName,pars->modName);
	if (funcptr != 0)
	{
		//TODO: here is not the correct place - change this 
		inter_module_put(pars->funcName);
	}
	else {
		res=FUNC_NOT_FOUND;
#ifdef TTCB
		*(long long*)(&tsout.buffer)=res;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else	//TTCB					
		rtf_put(fifo,&res,sizeof(res));		
#endif	//TTCB	
		return;
	}
	wcetptr=(wcetFunctype)inter_module_get_request(pars->wcetName, pars->modName);
			
	if (wcetptr != 0)
	{
		inter_module_put(pars->wcetName);
	}
	else {
		res=WCET_FUNC_NOT_FOUND;
#ifdef TTCB
		*(long long*)(&tsout.buffer)=res;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else	//TTCB					
		rtf_put(fifo,&res,sizeof(res));		
#endif	//TTCB
		return;
	}
	
	// preencher entrada na tabela
	executions[desc].func=funcptr;
	executions[desc].fifo=fifo;
	
	executions[desc].wcet=wcetptr();
	executions[desc].delay=pars->delay;
	executions[desc].max_exec=pars->max_exec;
	executions[desc].start_ev=pars->start_ev;
	strcpy(executions[desc].TraceName,pars->TraceName);
#ifdef TTCB
	executions[desc].Rindex = Rindex;
	memcpy(&(executions[desc].in), in, sizeof(in_hdr_t));
#endif //TTCB 	
	
	ret = rt_task_init(&(executions[desc].task_id),execTask,desc,4000,USER_TASK_PRIORITY,1,0);	
	
	if (ret!=0)																			
	  {
	    res=NO_AVAIL_EXEC;
#ifdef TTCB
		*(long long*)(&tsout.buffer)=res;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else //TTCB				
   	    rtf_put(fifo,&res,sizeof(res));
#endif //TTCB
          } 
          
        rt_task_resume(&(executions[desc].task_id));    
}

#endif //LOCAL_TIMELY

/*
+---------------------------------------------------------------------------+
|                       DETECCAO DE FALHAS LOCAIS                           |**
+---------------------------------------------------------------------------+**
   ****************************************************************************
*/

#ifdef LOCAL_FAULT 

void localFailure(int arg)
{
	RTIME now;
	RTIME nanos_past_start_ev;
	int res;
	int desc=arg;
	unsigned long irqstate;
#ifdef BENCH	
	unsigned long long tor=localFD[desc].start_ev+localFD[desc].spec;
	long long diff;
	unsigned long long t;
#endif	
#ifdef TTCB
	int Rindex = localFD[desc].Rindex;
	tcb_service_out tsout;
	tcb_service_in * in = &(localFD[desc].in);
	tsout.header_out.error = 0;
#endif //TTCB
	
	res=desc;

#ifdef TTCB
	*(int*)&(tsout.buffer)=res;
	process_output(localFD[desc].fifo, (out_hdr_t *)&tsout, sizeof(tcb_service_out), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else				
	rtf_put(localFD[desc].fifo,&res,sizeof(res));			   	    	   	   
#endif
	
	irqstate = rt_global_save_flags_and_cli();		
	
	//instante actual
	now=rt_get_time_ns();	
	
	// nanosegundos desde o instante inicial
	nanos_past_start_ev=now-localFD[desc].start_ev;
	
	
	 if (localFD[desc].spec <= nanos_past_start_ev )
	   {
	    rt_global_restore_flags(irqstate);	
	    // a falha ja ocorreu !
#ifdef TTCB
	    rtf_get(localFD[desc].fifo,&tsout,sizeof(tcb_service_out));	
#else				
	    rtf_get(localFD[desc].fifo,&res,sizeof(res));	
#endif
	    res=ALREADY_FAILED;
#ifdef TTCB
  	    *(int*)(&tsout.buffer)=res;
	    process_output(localFD[desc].fifo, (out_hdr_t *)&tsout, sizeof(tcb_service_out), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else				
	    rtf_put(localFD[desc].fifo,&res,sizeof(res));		
#endif
	    localFD[desc].func=0;
	    rt_task_delete(rt_whoami());	    
	   }
	  else
	  {	   
	   rt_task_make_periodic(rt_whoami(),nano2count(localFD[desc].start_ev+localFD[desc].spec),0);
	   rt_global_restore_flags(irqstate);
	   rt_task_wait_period();
	   
	   // neste ponto pode ser parada por endTCBlocalFD!!!!

#ifdef BENCH
	t=rt_get_time_ns();
	diff=t-tor;
	if (diff>max_release_delay_of_handler_LFD)
		max_release_delay_of_handler_LFD=diff;
#endif	
	   localFD[desc].faulty=1;	   
	   localFD[desc].func(desc);	   	   
	  } 
	
	rt_task_delete(rt_whoami());	    
}

/*
+----------------------------+
| inicia deteccao local      |
+----------------------------+
| retorna a aplicacao        |
|	   descriptor        | 
|	   FUNC_NOT_FOUND    |		
|	   NO_AVAIL_LOCAL_FD |
|	   ALREADY_FAILED    | 
+----------------------------+	   
*/

#ifdef TTCB
void TCBstartlocalFD(int fifo, tcb_service_in *in)  
#else		
void TCBstartlocalFD(int fifo,localFDPars* pars)
#endif
{	
	LFDhandtype     funcptr;
	wcetFunctype wcetptr;
	int	     res,ret;
	int a,desc=NO_AVAIL_LOCAL_FD;
#ifdef TTCB
	localFDPars *pars;
	int Rindex;
	tcb_service_out tsout;
	Rindex=EIDtoINDEX(in->header_in.eid);
	pars = (localFDPars *) in->buffer;
	tsout.header_out.error=0;
#endif	

	
	// procurar descriptor livre
	for (a=0;((a<MAX_LOCAL_FD)&&(desc==NO_AVAIL_LOCAL_FD));a++)
	{
		if (localFD[a].func==0)
			desc=a;
	}
	
	if (desc == NO_AVAIL_LOCAL_FD)
	{
		res=NO_AVAIL_LOCAL_FD;
#ifdef TTCB
		*(int*)(&tsout.buffer)=res;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tcb_service_out), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else						
		rtf_put(fifo,&res,sizeof(res));		
#endif	
		return;
	}
	
	//obter endereco da funcao	
	funcptr=(LFDhandtype)inter_module_get_request(pars->funcName, pars->modName);
	
	if (funcptr != 0)
	{
		inter_module_put(pars->funcName);
	}
	else
	{
		res=FUNC_NOT_FOUND;
#ifdef TTCB
		*(int*)(&tsout.buffer)=res;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tcb_service_out), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else						
		rtf_put(fifo,&res,sizeof(res));		
#endif	
		return;
	}
	
	wcetptr=(wcetFunctype)inter_module_get_request(pars->wcetName, pars->modName);
	
	
	if (wcetptr != 0)
	{
		inter_module_put(pars->wcetName);
	}
	else
	{
		res=WCET_FUNC_NOT_FOUND;
#ifdef TTCB
		*(int*)(&tsout.buffer)=res;
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tcb_service_out), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else						
		rtf_put(fifo,&res,sizeof(res));		
#endif	
		return;
	}
		
	// preencher entrada na tabela
 	localFD[desc].func=funcptr;
	localFD[desc].fifo=fifo;
	localFD[desc].start_ev=pars->start_ev;
	localFD[desc].deadline=pars->deadline;
	localFD[desc].spec=pars->spec;	
	localFD[desc].faulty=0;
#ifdef TTCB
	localFD[desc].Rindex = Rindex;
	memcpy(&(localFD[desc].in), in, sizeof(in_hdr_t));
#endif		

	ret = rt_task_init(&(localFD[desc].task_id),localFailure,desc,4000,USER_TASK_PRIORITY,1,0);	

	if (ret!=0)																			
	  {
	    res=NO_AVAIL_LOCAL_FD;
#ifdef TTCB
	    *(int*)(&tsout.buffer)=res;
	    process_output(fifo, (out_hdr_t *)&tsout, sizeof(tcb_service_out), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else						
   	    rtf_put(fifo,&res,sizeof(res));
#endif   
          } 
          
        rt_task_resume(&(localFD[desc].task_id));  
	
}


/*
+--------------------------------+
| termina medicao local          |
+--------------------------------+
| returna a aplicacao            |  
|           OK                   |
|	    INVALID_LOCALFD_DESC |
+--------------------------------+
*/
#ifdef TTCB
void TCBend_localFD (int fifo, tcb_service_in *in)  
#else
void TCBend_localFD (int fifo,int* descriptor)
#endif
{
	endLocalFDResult res;
	RTIME actTime;	
	unsigned long irqstate;
#ifdef TTCB
	int  * descriptor;
	int Rindex;
	tcb_service_out tsout;
	Rindex=EIDtoINDEX(in->header_in.eid);
	descriptor = (int *) in->buffer;
	tsout.header_out.error=0;
#endif		

	/* descriptor invalido ??? */
	if ( (*descriptor<0) || (*descriptor>=MAX_LOCAL_FD) || (localFD[*descriptor].func==0) ) 
	{
		res.stat=INVALID_LOCALFD_DESC;
#ifdef TTCB
		memcpy(&(tsout.buffer), &res, sizeof(endLocalFDResult));
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else							
		rtf_put(fifo,&res,sizeof(res));
#endif	
		return;
	}
	
	actTime=rt_get_time_ns();

	// parar a deteccao
	irqstate = rt_global_save_flags_and_cli();		
	if (!localFD[*descriptor].faulty)   
	{
	 rt_task_delete(&(localFD[*descriptor].task_id));
	}
	rt_global_restore_flags(irqstate);				
	
	// obter numero de nanosegundos desde o inicio da medicao
	res.procdelay=actTime-localFD[*descriptor].start_ev;
	
	res.end_ev=actTime;
	res.stat=OK;
	res.faulty=localFD[*descriptor].faulty;

#ifdef TTCB
		memcpy(&(tsout.buffer), &res, sizeof(endLocalFDResult));
		process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else	
	rtf_put(fifo,&res,sizeof(res));
#endif
	
	//libertar descriptor
	localFD[*descriptor].func=0;	
}
#endif // LOCAL_FAULT


/*
+------------------------------------+
| inicializa as variaveis do sistema | 
+------------------------------------+ 
*/

void initVariables(void)
{
	unsigned int a;
	
#ifdef DIST_FAULT
	for (a=0;a<65536;a++)
	   port_to_fifo[a]=-1;	
	
	for (a=0;a<MAX_SEND_TABLE_SIZE;a++)
	 send_info[a].is_free=1;
	 
	for (a=0;a<MAX_RCV_PLOAD_TABLE_SIZE;a++)
	 receive_payload_info[a].is_free=1;
	 
	for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++)
	 wait_infos[a].is_free=1; 	 

	for (a=0;a<MAX_RCV_TABLE_SIZE;a++)
	 receive_info[a].is_free=1; 
#endif //DIST_FAULT
 
	for(a=0;a<MAX_TCB_CONNECTIONS;a++)
		usedFifos[a]=false;

#ifdef LOCAL_MEASURE		
	for (a=0;a<MAX_LOCAL_MEASURES;a++)		
	    	localMeasures[a]=0;
#endif //LOCAL_MEASURE

#ifdef LOCAL_TIMELY	    	
	for (a=0;a<MAX_EXEC;a++)		
 	    	executions[a].func=0;    
#endif //LOCAL_TIMELY
	    	
#ifdef LOCAL_FAULT
	for (a=0;a<MAX_LOCAL_FD;a++)		
	    	localFD[a].func=0;    	    	
#endif //LOCAL_FAULT

#ifdef DIST_FAULT
	for (a=0;a<NUMBER_OF_MACHINES;a++)
		number_of_detections[a]=0;
	
	seq_numbers=1;    		    		    		
	n_sends_in_this_slot=0;
#endif //DIST_FAULT

#ifdef DISTRIBUTED
	for (a=0;a<NUMBER_OF_MACHINES;a++){
		gclock_info[a].vclock=0;
		gclock_info[a].error=0xFFFFFFFF;
	}	
#endif

#ifdef TTCB
	initVariablesTTCB();
#endif
}


/*
+-----------------------------------------------------+
| Cria os Fifos RT usados pela tcb                    | 
| 						      | 
| RETORNA: 0 se criou correctamente todos os fifos ou | 
|          -1 se falhou na criacao de algum fifo.     |  				
+-----------------------------------------------------+ 
*/

int createFifos(void)
{
	int cnt;
	
	printk("\n-creating the rt-fifos for ttcb/applications comunications\n\n");	

	
	printk("     .ttcb debug fifo (0) size=%d ",SIZEOF_DEBUG_FIFO);
	if (rtf_create(0, SIZEOF_TRACING_FIFO))
	{	
		printk("->failed\n");	
		printk("\naborting!!!\nTTCB not started\n\n\n");
		return -1;		
	}
	else
	  printk("->suceeded\n");
	

	// usado para requerer conexoes a tcb
	printk("     .ttcb request connection fifo (%d) size=%d ",TCB_REQ_FIFO,SIZEOF_FIFOS);
	if (rtf_create(TCB_REQ_FIFO, SIZEOF_FIFOS))
	{	
		printk("->failed\n");	
		printk("\naborting!!!\nTTCB not started\n\n\n");
		return -1;		
	}
	else
	  printk("->succeded\n");	


	  	
	// fifos disponiveis para as ligacoes
	for (cnt=0;cnt<MAX_TCB_CONNECTIONS;cnt++)
	{
		printk("     .ttcb connection number %d, fifo (%d) size=%d ",cnt+1,cnt+TCB_CON_STARTFIFO,SIZEOF_FIFOS);
		if (rtf_create(cnt+TCB_CON_STARTFIFO, SIZEOF_FIFOS))
		{	
			printk("->failed\n");	
			printk("\naborting!!!\nTTCB not started\n\n\n");
			return -1;		
		}
		else
		  	printk("->succeded\n");	
	}
	
	for (cnt=0;cnt<MAX_TCB_CONNECTIONS;cnt++)
	{
		printk("     .ttcb connection number %d (WaitInfo channel),fifo (%d) size=%d ",cnt+1,cnt+TCB_CON_STARTFIFO+INC_TO_WAIT_FIFO,SIZEOF_WAIT_INFO_FIFOS);
		if (rtf_create(cnt+TCB_CON_STARTFIFO+INC_TO_WAIT_FIFO, SIZEOF_WAIT_INFO_FIFOS))
		{	
			printk("->failed\n");	
			printk("\naborting!!!\nTTCB not started\n\n\n");
			return -1;		
		}
		else
		  	printk("->succeded\n");	
	}
	  
	return 0;
}


/*
+-------------------------------------+
| Elimina os Fifos RT usados pela tcb | 
+-------------------------------------+ 
*/

void destroyFifos(void)
{
	int cnt;

	rtf_destroy(0);	//tracing FIFO

	rtf_destroy(TCB_REQ_FIFO);
	
	for (cnt=0;cnt<MAX_TCB_CONNECTIONS;cnt++)
		{
		 rtf_destroy(cnt+TCB_CON_STARTFIFO);	
		 rtf_destroy(cnt+TCB_CON_STARTFIFO+INC_TO_WAIT_FIFO);	
		}
}

/*
+----------------------------------------------+
| Liberta os recursos alocados por uma conexao | 
+----------------------------------------------+ 
*/

void free_resources(int fifo)
{
	int a;
	unsigned long irqstate;
 	
	a=0; irqstate=0;
#ifdef LOCAL_TIMELY
 	//libertar execucoes atempadas pendentes            
     	for (a=0;a<MAX_EXEC;a++)
     	{
     		irqstate = rt_global_save_flags_and_cli();
     		if ((executions[a].func!=0)&&(executions[a].fifo==fifo))
     		{
		 rt_task_delete(&(executions[a].task_id));
		 executions[a].func=0; 		 
     		}
     		rt_global_restore_flags(irqstate);
     	}
#endif //LOCAL_TIMELY
     
#ifdef LOCAL_FAULT
        //libertar detecoes de falhas locais pendentes        
        for (a=0;a<MAX_LOCAL_FD;a++)
        {
        	irqstate = rt_global_save_flags_and_cli();
        	if ((localFD[a].func!=0)&&(localFD[a].fifo==fifo))
        	{ 		
 		if (!localFD[a].faulty)   
		 rt_task_delete(&(localFD[a].task_id));
		localFD[a].func=0; 		
        	}
        	rt_global_restore_flags(irqstate);
        }        
        
#endif //LOCAL_FAULT

#ifdef DIST_FAULT
        //deteccoes distribuidas em execucao
        //basta limpar a tabela de winfo - evitar que o handler dispare
        
        for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++)
        {
 		irqstate = rt_global_save_flags_and_cli();
 		if ( (wait_infos[a].is_free==0)&&(wait_infos[a].owner!=-1)&&(wait_infos[a].owner==fifo)){
 		  if (wait_infos[a].existsTask!=0)
 		  {
 		    rt_task_delete(&(wait_infos[a].rt_task_id));
 	    	    wait_infos[a].existsTask = 0; 	
 		  }   					     	  	
 		}
 		rt_global_restore_flags(irqstate);        	
        }
#endif //DIST_FAULT                                
}

#ifdef TTCB
void wait_info_to_fifo (int fifo, tcb_service_in *in)
{
	int Rindex;
	tcb_service_out tsout;
	Rindex=EIDtoINDEX(in->header_in.eid);
	tsout.header_out.error=0;

	process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);		
}

void stop_winfo_to_fifo (int fifo, tcb_service_in *in)
{
	int Rindex;
	tcb_service_out tsout;
	Rindex=EIDtoINDEX(in->header_in.eid);
	tsout.header_out.error=0;

	process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);		
}

#endif //TTCB

#ifdef DIST_FAULT
#ifdef TTCB
void msg_rcv_on_payload (int fifo, tcb_service_in *in);  
#else
void msg_rcv_on_payload(int fifo,msg_rcv_on_payloadPars* pars);
#endif //TTCB
#endif //DIST_FAULT

/*
+---------------------------------+
| recebe pedidos de servico a tcb | 
+---------------------------------+ 
*/

int serviceHandler(unsigned int fifo)
{
	ReqTcbHeader reqHead;
	char buff[SIZEOF_FIFOS-sizeof(ReqTcbHeader)];
#ifdef TTCB
	tcb_service_in * sv_in;
#endif

#ifdef DIST_FAULT
	int * p=NULL;
#endif	//DIST_FAULT

	if (rtf_get(fifo,&reqHead,sizeof(ReqTcbHeader))==sizeof(ReqTcbHeader))
	 {
	   //To avoid buffer overflows, we must check the size
	   if (reqHead.sizeofmsg > SIZEOF_FIFOS - sizeof(ReqTcbHeader)) 
	     reqHead.sizeofmsg = SIZEOF_FIFOS - sizeof(ReqTcbHeader);
	   if (reqHead.sizeofmsg!=0)
	     rtf_get(fifo,&buff,reqHead.sizeofmsg);	
	   

	 	switch (reqHead.srvc)
	 	{
			case get_connection: //only to avoid compilation warning
					   break;
	 		// fechar conexao
	 		case end_connection:
#ifndef TTCB
	 				    free_resources(fifo);
	 				    usedFifos[fifo-TCB_CON_STARTFIFO]=false;  				    		
#endif
	 				    break;
	 		//obter info. do kernel
			case get_kernel_info:
#ifdef TTCB
		     		process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
	   			TCBget_kernel_info(fifo, (tcb_service_in *)&buff);
#else							
				TCBget_kernel_info(fifo);		
#endif		
				break;	

	 		// obter estampilha temporal
	 		case get_timestamp:
#ifdef TTCB
		     		process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
	   			TCBget_timestamp(fifo, (tcb_service_in *)&buff);
#else
			        TCBget_timestamp(fifo);
#endif
	 		        break;

#ifdef DISTRIBUTED
		 	// obter estampilha temporal global
	 		case get_global_timestamp:
#ifdef TTCB
		     		process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
	   			TCBget_global_timestamp(fifo, (tcb_service_in *)&buff);
#else
  			        TCBget_global_timestamp(fifo);
#endif //TTCB
	 		        break;	   	 			
			// obter estampilha temporal local/global
	 		case get_local_global_timestamp:
#ifdef TTCB
		     		process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
	   			TCBget_global_local_timestamp(fifo, (tcb_service_in *)&buff);
#else
			        TCBget_global_local_timestamp(fifo);
#endif
	 		        break;
	 	
#endif //DISTRIBUTED			   
	 		// obter id da TCB
	 		case get_id: 		
#ifdef TTCB
		     		process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
	   			TCBget_id(fifo, (tcb_service_in *)&buff);
#else
				TCBget_id(fifo);
#endif		
				break;		   

#ifdef LOCAL_MEASURE
	 				   
	 		//comecar medicao local
			case start_localM: 

#ifdef TTCB
			  	process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
			  	TCBstart_local_measure(fifo,(tcb_service_in*)&buff);
#else // TTCB
			   	TCBstart_local_measure(fifo,(RTIME*)&buff);
#endif // TTCB 


				break;

					   
			// terminar medicao local					   
			case end_localM	 : 
#ifdef TTCB
			  	process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
			  	TCBend_local_measure(fifo,(tcb_service_in*)&buff);
#else // TTCB	
				TCBend_local_measure(fifo,(int*)&buff);
#endif // TTCB		
				break;  		 				   

#endif // LOCAL_MEASURE
					   
#ifdef LOCAL_TIMELY
			// execucao atempada		   
			case exec	 : 
#ifdef TTCB
			  	process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
			  	TCBexec(fifo,(tcb_service_in*)&buff);
#else		//TTCB
				TCBexec(fifo,(execPars*)&buff);
#endif		//TTCB
				break;			   
			
#endif // LOCAL_TIMELY

#ifdef LOCAL_FAULT
			case start_localFD: 
#ifdef TTCB
			  	process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
			  	TCBstartlocalFD(fifo,(tcb_service_in*)&buff);
#else	//TTCB
				TCBstartlocalFD(fifo,(localFDPars*)&buff);					   
#endif		//TTCB
				break;
		        
		        // terminar deteccao local					   
			case end_localFD : 
#ifdef TTCB
			  	process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
			  	TCBend_localFD(fifo,(tcb_service_in*)&buff);
#else	//TTCB
	  	  	        TCBend_localFD(fifo,(int*)&buff);
#endif		//TTCB
				break;			   
			
#endif // LOCAL_FAULT

#ifdef DIST_FAULT
			case start_distributedFD: 
#ifdef TTCB
			  	process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
			  	TCBstart_distributedFD(fifo,(tcb_service_in*)&buff);
#else		//TTCB
				TCBstart_distributedFD(fifo,(distributedFDPars*)&buff);
#endif		//TTCB
			   	break;
			case msg_received_on_payload: 
#ifdef TTCB
			  	process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));
			  	msg_rcv_on_payload(fifo,(tcb_service_in*)&buff);
#else	
				msg_rcv_on_payload(fifo,(msg_rcv_on_payloadPars*)&buff);
#endif		
        		        break;		   	  			    
			case winfo_to_fifo: 
#ifdef TTCB
			  	process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));	
				sv_in = (tcb_service_in*)&buff;
				p=(unsigned int*)&(sv_in->buffer);
				wait_info_to_fifo(fifo,(tcb_service_in*)&buff);
#else
				p=(unsigned int*)&buff;
				rtf_put(fifo,(int*)p,sizeof(int));
#endif	
				port_to_fifo[*p]=fifo+INC_TO_WAIT_FIFO;
			break;
		
			case stop_winfo_port:
#ifdef TTCB
			  	process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(tcb_service_out));	
				sv_in = (tcb_service_in*)&buff;
				p=(unsigned int*)&(sv_in->buffer);
				stop_winfo_to_fifo(fifo,(tcb_service_in*)&buff);
#else	
				p=(unsigned int*)&buff;
				rtf_put(fifo,(int*)p,sizeof(int));
#endif	

			        port_to_fifo[*p]=-1;
			        break;		    		   	 					
#endif // DIST_FAULT	   	 

#ifdef TTCB

			case localAuthentication_enum:
				TTCBlocalAuthentication(fifo, (char *)&buff,reqHead.sizeofmsg); 
				break;

	 		case end_connection_TTCB_enum:
				process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(error_out_t));
				TTCBend_connection(fifo, (eid_in_t *)&buff);
	 			break;

#ifdef DIST_AGREEMENT
			case propose_enum:
			       //temp_time0 = get_vclock();
				process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(propose_out_t));
				TTCBpropose(fifo, (propose_in_t *)&buff); 
				break;

			case decide_enum:
			        //temp_time2 = get_vclock();
				process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(decide_out_t));
				TTCBdecide(fifo, (decide_in_t *)&buff); 
				break;
#endif

#ifdef LOCAL_RANDOM
			case getRandom_enum:
				process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(rand_out_t));
				TTCBrandom(fifo, (eid_in_t *)&buff); 
				break;
#endif
			case getLocalPublicKey_enum:
			        TTCBGivePublicKey(fifo);
			        break;
				
			case getLocalTTCBid_enum:
				break;

			case crashedLocalTTCBid_enum:
				process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(error_out_t));
				TTCBcrashedLocalTTCBid(fifo, (localTTCBid_in_t *) &buff);
				break;

			case crashedLocalTTCBeid_enum: 
			        process_input(fifo, reqHead, (in_hdr_t *)&buff, sizeof(error_out_t));
				TTCBcrashedLocalTTCBid(fifo, (localTTCBid_in_t *) &buff);
				break;			

#endif		
		        default: 
			        DEBUG_MSG(SERVICE_LEVEL,"TTCB called with invalid function (%i)!\n",reqHead.srvc);				
				break;  
	 	}
	 }	
	
	return 0;
}



/*
+-----------------------------------------------------+
|  devolve um numero de conexao livre, marca como     | 
|  ocupada, e associa handler que trata os pedidos da | 
|  conexao criada.                                    | 
| se nao existir devolve NO_AVAIL_CONN                | 	 
+-----------------------------------------------------+ 
 */

int requestfreeConnection(void)
{
	int a;
	
	for (a=0;a<MAX_TCB_CONNECTIONS;a++)
	{
	  if (!usedFifos[a])
	  {
	   usedFifos[a]=true;
	   rtf_create_handler(a+TCB_CON_STARTFIFO, &serviceHandler);
	   return (a+TCB_CON_STARTFIFO);
	  } 
	}
	return NO_AVAIL_CONN;
}



/*
+-----------------------------------------------------+
| handler invocado quando ha actividade no fifo usado | 
| para invocar conexoes a tcb                         | 	 
+-----------------------------------------------------+ 
*/

int conHandler(unsigned int fifo)
{
	ReqTcbHeader reqHead;
	int con;
#ifndef DISTRIBUTED
	unsigned char state_of_operation=NORMAL_STATE;
#endif

	if (rtf_get(TCB_REQ_FIFO,&reqHead,sizeof(ReqTcbHeader))==sizeof(ReqTcbHeader))
	{
		if (reqHead.srvc==get_connection)
		{  
			if (state_of_operation!=NORMAL_STATE)
			{
			 con=IN_INIT_STATE;	
			 rtf_put(TCB_REQ_FIFO,&con,sizeof(int));
			} 
			else
			 {
			  // verifica conexao livre, NO_AVAIL_CONN se nao existe	
			  con=requestfreeConnection();
			  rtf_put(TCB_REQ_FIFO,&con,sizeof(int));
			 } 
		}
	}	

	return 0;
}

/*
+-------------+
| DEBUG TOOLS | ***********************************************************
+-------------+ 
*/

//faz um dump do estado interno da tcb   <debug/dump_state.c>
//TODO: continuar!!!!!!!!!!!!!!!!!!!!!!!
void dump_state(void){
	int a;

	rt_printk("---------------------\n");
	rt_printk(" TTCB internal state \n");
	rt_printk("---------------------\n\n");

#ifdef DISTRIBUTED
	rt_printk("1) System View:\n");
	for (a=0;a<NUMBER_OF_MACHINES;a++){
		if (system_view[a]==NOT_PRESENT)
			rt_printk("   machine %u, %s\n",a,"Not Present");
		if (system_view[a]==IN_INIT)
			rt_printk("   machine %u, %s\n",a,"In Init");	
		if (system_view[a]==NORMAL)
			rt_printk("   machine %u, %s\n",a,"Active");
	}
	
	rt_printk("2) Vclock = %llu nanos\n",get_vclock());
#endif

#ifdef DIST_FAULT
	rt_printk("3) Distributed Failure Detection Service:\n");
	rt_printk("   .number of active detections on each machine\n");
	for (a=0;a<NUMBER_OF_MACHINES;a++){
		rt_printk("      machine %u, %u detections\n",a,number_of_detections[a]);
	}
	rt_printk("   .wait_info table:\n");
	for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++){
		if (!wait_infos[a].is_free){
			rt_printk("      position %u\n",a);
			rt_printk("        to <%u>, from <%u>, seq <%u>\n",wait_infos[a].to,wait_infos[a].from,wait_infos[a].seq);
		}
	}
#endif //DIST_FAULT
}


//mostra os resultados do bench   <bench/bench_results.c>
void bench_results(void){

#ifdef BENCH		
	rt_printk("--------------------------\n");
	rt_printk(" TTCB benchmarking results \n");
	rt_printk("--------------------------\n\n");	

#ifdef DISTRIBUTED			
	rt_printk("Maximum execution time of read task = %u nanos\n",bench_max_exec_read);		
	
	rt_printk("Maximum execution time of send task = %u nanos\n",bench_max_exec_send);		
	rt_printk("Maximum roundtrip time = %u us\n",max_roundtrip);
	rt_printk("Maximum release time for the send task = %u nanos\n",max_release_time_send);	
	rt_printk("Maximum release time for the read task = %u nanos\n",max_release_time_read);	
	rt_printk("Global clock precision = %u nanos\n",max_global_clock_drift);	
#endif
#ifdef DIST_FAULT
	rt_printk("Maximum latency of a distributed failure detection = %u nanos\n",max_release_delay_of_handler_DFD);	
#endif
#ifdef LOCAL_FAULT
	rt_printk("Maximum latency of a local failure detection = %u nanos\n",max_release_delay_of_handler_LFD);	
#endif	
#else//BENCH
	rt_printk("The TTCB is not running on benchmarking mode\n");
#endif 
}

int debugHandler(unsigned int fifo)
{
	int command;

	if (rtf_get(0,&command,sizeof(command))==sizeof(command))
	{
//FALTA PROTEGER concorrencia !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//LIMITAR ACESSOS P/segundo
		switch (command) {	
		  case DUMP_INTERNAL_STATE: 
	  					dump_state();
						break;
	  	  case BENCH_RESULTS: 		bench_results();
						break;
		}
	}

	return 0;
}

/* ******************************************************************* */

int setconHandler(void)
{

	printk("\n-installing ttcb debug handler on fifo (%d) ",0);

	if (rtf_create_handler(0, &debugHandler))
	{
		printk("->failed\n");	
		printk("\naborting!!!\nTTCB not started\n\n\n");
		return -1;			
	}
	else
		printk("->succeded\n");	

	printk("\n-installing ttcb connection handler on fifo (%d) ",TCB_REQ_FIFO);

	if (rtf_create_handler(TCB_REQ_FIFO, &conHandler))
	{
		printk("->failed\n");	
		printk("\naborting!!!\nTTCB not started\n\n\n");
		return -1;			
	}
	else
		printk("->succeded\n");	
		
	return 0;	
		
}

/* *********************************************************************** */


#ifdef DIST_FAULT

/*
+-----------------------------------------------------------------+
| Verifica se a informacao de falha relativa a uma dada mensagem  |
| ja chegou (no canal de controlo). Processa a analise temporal da|
| da mensagem.							  |			
+-----------------------------------------------------------------+ 
*/

int is_already_in_rcvInfo(msg_rcv_on_payloadPars* pars) 
{
	int a;
	int ret=0;
	unsigned long h;
	unsigned long long t;
	
	
	for (a=0;a<MAX_RCV_TABLE_SIZE;a++)
	{
		if (receive_info[a].is_free==0)
		 if ((pars->from==receive_info[a].from)&&(pars->seq==receive_info[a].seq))
		 {
		   // ja falhou	
		   if (pars->time>receive_info[a].time_of_failure)	
		    {
		    //rt_printk("Ja ca tenho e falhou\n");
		    	receive_info[a].fail=1;
		    	receive_info[a].lateness_available=1;
		    	h=pars->time-receive_info[a].time_of_failure;
		    	h=h/1000;
		    	receive_info[a].lateness_degree=(unsigned long)h;		    	
		    }
		     else  // nao falhou 
		      {
		      //rt_printk("Ja ca tenho e nao falhou\n");
		      	receive_info[a].fail=0;
		      	t=receive_info[a].time_of_failure-pars->time;
		      	if (t<268435456000)  //max: 2^28 us = 2^28*1000 nanos
		      	{
		      	  receive_info[a].lateness_available=1;
		      	  h=(unsigned long)t;
		    	  h=h/1000;
		    	  receive_info[a].lateness_degree=(unsigned long)h;		    	
		    	}
		    	 else
		    	  receive_info[a].lateness_available=0;  
		      }
		    ret=1;	 	
		 }
	}
	return ret;
}

#endif // DIST_FAULT

#ifdef DIST_FAULT
/*
+-----------------------------------------------------------------+
| Esta funcao e chamada pelas aplicacoes quando recebem pacotes   |
| no canal canal de payload.					  | 
+-----------------------------------------------------------------+ 
*/

#ifdef TTCB
void msg_rcv_on_payload (int fifo, tcb_service_in *in)  
#else
void msg_rcv_on_payload(int fifo,msg_rcv_on_payloadPars* pars)
#endif
{	
	int c;
	int res;
	unsigned long long act_time;
#ifdef TTCB
	msg_rcv_on_payloadPars* pars;
	int Rindex;
	tcb_service_out tsout;
	Rindex=EIDtoINDEX(in->header_in.eid);
	pars = (msg_rcv_on_payloadPars*)in->buffer;
	tsout.header_out.error=0;
#endif
	
	c=0;
	TCBenter_in_critical_section();
			
	if (!is_already_in_rcvInfo(pars))
	{
			
	act_time=rt_get_time_ns();
	
	while (c<MAX_RCV_PLOAD_TABLE_SIZE)			
	{
	 if (receive_payload_info[c].is_free)
	  break;
	 //Garbage Collector
	 //Quando ocorrem paragens nos nodos da TCB, podem existir mensagens
	 //que sao recebidas no canal de payload, e cuja informacao de falha
	 //nao e recebida no canal de controlo (o nodo emissor falha antes de
	 //enviar a informacao que tem na tabela send_info.) 
	 if (act_time>receive_payload_info[c].clean_up_after)
	  break;
	 c++;
	}
	
	if (c!=MAX_RCV_PLOAD_TABLE_SIZE)
		{
			receive_payload_info[c].is_free=0;
			//valor seguro
//TODO:	ver tempo seguro!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			receive_payload_info[c].clean_up_after=pars->time+(2*((SEND_PERIOD*(NETWORK_OMISSION_DEGREE+1))+READ_PERIOD));  
			receive_payload_info[c].time_of_receipt=pars->time;
			receive_payload_info[c].seq=pars->seq;
			receive_payload_info[c].from=pars->from;
			TCBleave_critical_section();
			res=0;
#ifdef TTCB
			*((int*)(&tsout.buffer))=res;
			process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else		
		  	rtf_put(fifo,&res,sizeof(res));
#endif	  //TTCB
		}	
		 else   // nao ha espaco na tabela => simular uma omissao no canal de payload
		  {
	//	        rt_printk("No Space on rcvpl....omission simulated\n");
		  	TCBleave_critical_section();
		  	res=1;
#ifdef TTCB
			*((int*)(&tsout.buffer))=res;
			process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else		  
		  	rtf_put(fifo,&res,sizeof(res));
#endif	 //TTCB 
		  }
	}	  
       else
        {
           TCBleave_critical_section();	
           res=0;           
#ifdef TTCB
	   *((int*)(&tsout.buffer))=res;
	   process_output(fifo, (out_hdr_t *)&tsout, sizeof(tsout), in->header_in.eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
#else	           
	   rtf_put(fifo,&res,sizeof(res));	   
#endif
        }		  
}

#endif //DIST_FAULT

#ifdef DIST_FAULT

/*
+-----------------------------------------------------------------+
| Verifica se a mensagem recebida de (from) com a sequencia (seq) | 
| ja foi recebida. Em caso afirmativo, verifica se foi pontual ou |
| atrasada.            						  |	
+-----------------------------------------------------------------+ 
*/

void already_received_on_payload(int a,unsigned char from,unsigned long seq,unsigned long long time_of_failure)
{
	int c;
	unsigned long t;
	unsigned long long l;
	
	for (c=0;c<MAX_RCV_PLOAD_TABLE_SIZE;c++)
	{
		if (receive_payload_info[c].is_free==0)
		{
			// ja foi recebida?	
			if ((receive_payload_info[c].from==from)&&(receive_payload_info[c].seq==seq))
			{
				// neste caso chegou atrasada
				if (receive_payload_info[c].time_of_receipt>time_of_failure)
				{
					receive_info[a].fail=1;
					receive_info[a].lateness_available=1;
					// calcular o grau de atraso
					t=receive_payload_info[c].time_of_receipt-time_of_failure;
					t=t/1000;
					receive_info[a].lateness_degree=t;		
				}
				 else   // chegou pontualmente
				  {
				  	receive_info[a].fail=0;
					l=time_of_failure-receive_payload_info[c].time_of_receipt;
					//ver se conseguimos colocar o grau de atraso nos 29 bits reservados
				      	if (l<268435456000)  //max: 2^28 us = 2^28*1000 nanos
				      	{
		      	  		  receive_info[a].lateness_available=1;
		      	  		  t=(unsigned long)l;
		    	  		  t=t/1000;
		    	  		  receive_info[a].lateness_degree=t;		    	
		    			}
		    	 		else  // muito grande!!
		    	  		  receive_info[a].lateness_available=0;  			  	
				  }		
			  receive_payload_info[c].is_free=1;  	  
			  return ;
			}
		}
	}
}
#endif //DIST_FAULT

#ifdef DIST_FAULT
inline int get_wait_info_pos(unsigned long seq,unsigned char to)
{
	int c;
	for (c=0;c<MAX_WAIT_INFO_TABLE_SIZE;c++)
	{
	 	if ((wait_infos[c].is_free==0)&&(wait_infos[c].seq==seq)&&(wait_infos[c].from==to))
	 	  return c;
	 	 
	}
	return -1;
}
#endif

/*
+-----------------------------------------------------+
|  Envia a informacao relativa a uma dada mensagem    | 
|  as aplicacoes que esperam essa mesma informacao.   | 
|  send_fifo!=-1 significa que o emissor pertence ao  | 
|      ao nodo e tb devera receber a informacao.      | 	 
+-----------------------------------------------------+ 
*/

#ifdef DIST_FAULT

void reply_wInfo(int pos_in_wi,int send_fifo,unsigned int dest_port)
{
	fail_info_to_app res;

	memcpy(&res.failure_info[0],&wait_infos[pos_in_wi].failure_info[0],sizeof(fail_info)*NUMBER_OF_MACHINES);
	res.from=wait_infos[pos_in_wi].from;
	res.seq=wait_infos[pos_in_wi].seq;
	res.port=wait_infos[pos_in_wi].destination_port;
	
	if (send_fifo!=-1){
	  rtf_put(send_fifo,&res,sizeof(res));
	}

	if (port_to_fifo[dest_port]!=-1){
		 rtf_put(port_to_fifo[dest_port],&res,sizeof(res));
	 }
	
	if ((wait_infos[pos_in_wi].existsTask!=0)&&(wait_infos[pos_in_wi].any_fail==0)){
		rt_task_delete(&(wait_infos[pos_in_wi].rt_task_id));
		wait_infos[pos_in_wi].is_free=1;
	}
	else
	if ((wait_infos[pos_in_wi].existsTask!=0)&&(wait_infos[pos_in_wi].any_fail)){
		// a task quando acordar deve libertar a entrada na tabela
		wait_infos[pos_in_wi].any_fail = 2;	
	}
	else{
	  wait_infos[pos_in_wi].is_free=1;
	}
}
#endif //DIST_FAULT



/*
+-----------------------------------------------------+
|  PROCESSA UM PEDIDO DE DETECAO DE FALHA   	      | 
+-----------------------------------------------------+ 
*/

#ifdef DIST_FAULT
void process_request(unsigned char from,unsigned char to,unsigned long seq,long time_to_failure,unsigned long RndTripVal,unsigned long long time_of_receipt,unsigned int dest_port)
{												
	int a,c;

	unsigned long long time_of_failure;
	char to_array[NUMBER_OF_MACHINES];		
	unsigned char mask=1,dest=to;
	unsigned char number_of_dest=0;
	unsigned int pos_in_wInfo=0;
		
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	{
		if (mask&dest)
		{
		  to_array[a]=1;
		  number_of_dest++;
		}  
		   else
		    to_array[a]=0;
		mask<<=1;
	}
	
	// nao e destinada a esta tcb entao nao faz nada
	if (!to_array[tcb_machine_id]){
	 DEBUG_MSG(TFD_LEVEL,"TFD Request: from <%u>, to <%u>, seq <%u>, dest_port <%u> ----> IGNORED \n",from, to, seq, dest_port);
	 return;
	}
	
	DEBUG_MSG(TFD_LEVEL,"TFD Request: from <%u>, to <%u>, seq <%u>, dest_port <%u>\n",from, to, seq, dest_port);

	// coloca na tabela de pedidos recebidos
	for (a=0;a<MAX_RCV_TABLE_SIZE;a++)
	{
		if (receive_info[a].is_free)
		 {
		 	receive_info[a].is_free=0;
		 	// time_of_recipt contem e a estampilha temporal
		 	// do instante em que a mensagem foi recebida.	
		 	time_of_failure=time_of_receipt+((long long)time_to_failure*1000)-((unsigned long long)RndTripVal*1000);		 			 	
 
		 	receive_info[a].time_of_failure=time_of_failure;
		 	receive_info[a].seq=seq;
		 	receive_info[a].from=from;
		 	receive_info[a].to=to;
		 	// enquanto nao se receber atempadamente assume-se	
		 	// que falhou
		 	receive_info[a].fail=1;		 			 	
		 	receive_info[a].destination_port=dest_port;	 	
		 	receive_info[a].lateness_degree=0;
		 	receive_info[a].lateness_available=0;
		 	// verificar se a mensagem ja chegou  
			already_received_on_payload(a,from,seq,receive_info[a].time_of_failure);		 	
			
			// o emissor da mensagem ja criou a estrutura de Winfo
			if (from==tcb_machine_id) return ;
			
			// os destinatarios da mensagem deverao receber a 
			// informacao relativa ao servico de detecao de falhas
			
			for (c=0;c<MAX_WAIT_INFO_TABLE_SIZE;c++)
		 	{
		 		if (wait_infos[c].is_free)
		 		{
		 		 pos_in_wInfo=c;
		 		 break;
		 		} 
		 	}
		 	// tem de se parar o sistema pois ficaria 
		 	// incoerente.
		 	if (pos_in_wInfo==-1)
		 	 {
#ifdef RELEASE	 
	 			while (1){
#endif
		 	 		rt_printk("TTCB ERROR: No space in wInfo to process the request\n");
#ifdef RELEASE	 	 
		 	 	}
#endif	 
		 	 }
		 	for (c=0;c<NUMBER_OF_MACHINES;c++)
		 	  wait_infos[pos_in_wInfo].failure_info[c].fail=2;
		 	wait_infos[pos_in_wInfo].is_free=0;
		 	//como recebi um pedido, o emissor da mensagem e remoto 	
		 	wait_infos[pos_in_wInfo].send_fifo=-1; 
		 	wait_infos[pos_in_wInfo].owner=-1;
		 	wait_infos[pos_in_wInfo].n_dests=number_of_dest;  // numero de destinatarios da mensagem
		 	wait_infos[pos_in_wInfo].seq=seq;
		 	wait_infos[pos_in_wInfo].any_fail=0;
		 	wait_infos[pos_in_wInfo].existsTask=0; // a detecao atempada e no emissor
		 	wait_infos[pos_in_wInfo].to=to;
		 	wait_infos[pos_in_wInfo].from=from;
		 	wait_infos[pos_in_wInfo].destination_port=dest_port;
		 	wait_infos[pos_in_wInfo].failure_at_time=receive_info[a].time_of_failure;
			return;
		 }
	}	
	while (1){
		rt_printk("TTCB PANIC: receive table overflow\n");
	};   // nao devera acontecer
}
#endif

/*
+-----------------------------------------------------+
|  PROCESSA UMA RESPOSTA DE DETECAO DE FALHA   	      | 
+-----------------------------------------------------+ 
*/

#ifdef DIST_FAULT
void process_resp(unsigned char Bbit,unsigned char to,unsigned char from,unsigned long seq,unsigned long time_to_failure,unsigned int dest_port)
{
	unsigned long h;
	unsigned char fail,lateness_avail;
	unsigned long ldegree;
	int pos_in_wInfo;

	h=time_to_failure;
	fail=(h&0x40000000)>>30;
	
	h=time_to_failure;
	lateness_avail=(h&0x20000000)>>29;
	
	h=time_to_failure;
	ldegree=time_to_failure&0x1FFFFFFF;
	
	if (to==tcb_machine_id)  //e para mim
	  number_of_detections[from]--;

	pos_in_wInfo=get_wait_info_pos(seq,to);					
	if (pos_in_wInfo==-1)  //Bbit==1 but is not for me
	{
		//rt_printk("Not for me seq:%u to:%u BBit:%u\n",seq,to,Bbit);
		//for (c=0;c<MAX_WAIT_INFO_TABLE_SIZE;c++)
		//{
		//	printk("SEQ: %u  AV: %u --",wait_infos[c].seq,wait_infos[c].is_free);
		//}
		//while (1) {}
		return;
	}
	//recebi informacao da ocorrencia de uma falha e e a primeira
	if (fail&&!wait_infos[pos_in_wInfo].any_fail)
	{
		wait_infos[pos_in_wInfo].any_fail=1;	
		//sou o emissor?
		if (wait_infos[pos_in_wInfo].existsTask!=0)
		  //acordar a tarefa de deteccao atempada
		  //comeca quando a tarefa de leitura terminar	  	  
	  	  rt_task_resume(&(wait_infos[pos_in_wInfo].rt_task_id));		
	}
	wait_infos[pos_in_wInfo].failure_info[from].fail=fail;
	wait_infos[pos_in_wInfo].failure_info[from].lateness_avail=lateness_avail;
	if (lateness_avail)
	 wait_infos[pos_in_wInfo].failure_info[from].lateness_degree=ldegree;
	  else 
	   wait_infos[pos_in_wInfo].failure_info[from].lateness_degree=0;
	wait_infos[pos_in_wInfo].n_dests--;   
	if (wait_infos[pos_in_wInfo].n_dests==0) //we may reply now because no more info's will came
  	  reply_wInfo(pos_in_wInfo,wait_infos[pos_in_wInfo].send_fifo,dest_port);
		    
	DEBUG_MSG(TFD_LEVEL,"TFD Reply: fail <%u> lateness degree available <%u> to:<%u> from<%u> lateness degree<%u> seq <%u>\n",fail,lateness_avail,to,from,ldegree,seq);
}																						
#endif

/*
+-----------------------------------------------------+
| PROCESSA UMA MENSAGEM RECEBIDA NO CANAL DE CONTROLO | 
+-----------------------------------------------------+ 
*/
#if defined(DIST_FAULT) || defined (DIST_AGREEMENT)
void process_read(msgtr_hdr *hdr,unsigned long RndTripVal,unsigned long long time_of_receipt)
{
	info_field* pt=(info_field*)(hdr+1);			
#ifdef DIST_FAULT
	unsigned long h;
	unsigned char R,to;
	unsigned long seq,sign_bit;
	long time_to_failure;
	unsigned char Bbit;
#endif
	int items_read=0;
#ifdef DIST_AGREEMENT
	CSTable_field *cst_ptr;
	short Rindex, j;
	tag_t tag;
#endif		

#ifdef DIST_FAULT
	while ((pt->seq_to!=0x00000000)&&(items_read<max_fields)){

	 items_read++;		
	 h=pt->r_time_to_failure;
	 R=(h&0x80000000)>>31;   //R=Request bit
	 h=pt->r_time_to_failure;
	 if (R==0) 	         // pedido
	 {
	 sign_bit=(h&0x40000000);        //tempo ate falhar e positivo ou negativo? 	 
	 h=pt->r_time_to_failure;
	 time_to_failure=(h&0x3FFFFFFF);  
	 if (sign_bit)                    // e negativo 
	 {
	  time_to_failure*=-1;
	 } 
	}
	 else         
	  time_to_failure=(h&0x7FFFFFFF);  
	
	 h=pt->seq_to;
	 to=(h>>24);                     // enderecado a quem?
	 h=pt->seq_to;
	 seq=h&0x007FFFFF;		 // sequencia
	
	 h=pt->seq_to;
	 Bbit=((h&0x00800000)!=0);       // B bit activo especifica que
	 				 // todos os nodos devem processar a 
	 				 // a resposta		
			
	if (R==0){   // request
	  DEBUG_MSG(TFD_LEVEL,"Request: from <%u> to <%u>\n",hdr->machine_src_id,to);
	  process_request(hdr->machine_src_id,to,seq,time_to_failure,RndTripVal,time_of_receipt,pt->destination_port);				
	}
	else        // reply
	// so processa as mensagens destinadas a si; como o Bbit activo; ou enviadas por si
	if ((to==tcb_machine_id)||(Bbit)||(hdr->machine_src_id==tcb_machine_id)) 
	{
	  DEBUG_MSG(TFD_LEVEL,"Reply: from <%u> Bbit<%u>\n",hdr->machine_src_id,Bbit);	
	  process_resp(Bbit,to,hdr->machine_src_id,seq,time_to_failure,pt->destination_port);
	}
	pt++;
	}		
#endif

#ifdef DIST_AGREEMENT
#ifdef DIST_FAULT
	if (items_read<max_fields) //stopped because pt->seq_to==0x00000000 so pt is not
		pt++;			//yet pointing to Agreement header => increment
#endif	

	cst_ptr = (CSTable_field *)pt;

	items_read=0;
	while ((items_read < max_fields_agreement) && (cst_ptr->tstart != 0)) {
		if ((Rindex = search_CDTable(cst_ptr->elist, cst_ptr->tstart, cst_ptr->decision)) != -1) {
		  //rtl_printf("Received proposal to a known agreement: tstart=%lld\n", cst_ptr->tstart);
			j = index_eid_in_elist(cst_ptr->eid, cst_ptr->elist);
			CDTable[Rindex].vtable[j].is_free = 0;
			memcpy(CDTable[Rindex].vtable[j].value, cst_ptr->value, sizeof(value_t));
		} else {
		    //rtl_printf("Received proposal to a new agreement: tstart=%lld\n", cst_ptr->tstart);
			tag = get_tag();
			if ((Rindex = get_freeR_in_CDTable()) == -1) { //never happens if Access Control right...
				DEBUG_MSG(SERVICE_LEVEL,"No free space for Agreements!!\n");
				Rindex = 0;
			}
			CDTable[Rindex].tag = tag;
			memcpy(CDTable[Rindex].elist, cst_ptr->elist, sizeof(elist_t));
			CDTable[Rindex].tstart = cst_ptr->tstart;
			CDTable[Rindex].decision = cst_ptr->decision;
			j = index_eid_in_elist(cst_ptr->eid, cst_ptr->elist);
			CDTable[Rindex].vtable[j].is_free = 0;
			memcpy(CDTable[Rindex].vtable[j].value, cst_ptr->value, sizeof(value_t));
		}
		cst_ptr++;
		items_read++;
	}
#endif   //DIST_AGREEMENT

}
#endif //DIST_FAULT || DIST_AGREEMENT

/*
+-----------------------------------------------------+
| VERIFICA SE NAO RECEBEU INFO DE FALHA DEVIDO A UMA  |
| PARAGEM DE UM NODO.				      | 
+-----------------------------------------------------+ 
*/

#ifdef DIST_FAULT

void process_wInfos(void)
{	
	int c,a;
	char to_array[NUMBER_OF_MACHINES];		
	unsigned char mask;
	unsigned char dest;
	unsigned char can_delivery;
	
	for (c=0;c<MAX_WAIT_INFO_TABLE_SIZE;c++)
	{		
		// ja deveriamos ter recebido todas as respostas									
		// este valor e um lower bound - so para efeitos de optimizacao	
//VER ISTO BEM!!!!!!!!!!!!	
		if ((wait_infos[c].is_free==0)&&(rt_get_time_ns()>=(wait_infos[c].failure_at_time+SEND_PERIOD+2*MAXIMUM_ROUNDTRIP-MINIMUM_DELIVERY_DELAY))&&(wait_infos[c].any_fail != 2)) 
		{
		  mask=1;
		  dest=wait_infos[c].to;
		  
		  for (a=0;a<NUMBER_OF_MACHINES;a++)
		  {
		   if (mask&dest)
  		    to_array[a]=1;
		   else
		    to_array[a]=0;
		    mask<<=1;
		 }
		
		  can_delivery=1;	
		  for (a=0;a<NUMBER_OF_MACHINES;a++)
		  {
		     if (to_array[a]){
		     	//a maquina receptora esta como destinataria mas falhou
		     	//ainda nao recebi mas tb nao vou receber pq a maquina falhou
		  	if ((wait_infos[c].failure_info[a].fail==2)&&(system_view[a]!=NORMAL)) 
		  	{
		  	 if (wait_infos[c].send_fifo!=-1)		  	 
		  	   number_of_detections[a]--;	
	  
		  	 wait_infos[c].failure_info[a].fail=1;
	  		 wait_infos[c].n_dests--;
		  	 wait_infos[c].failure_info[a].lateness_avail=0;
		  	 if (!wait_infos[c].any_fail)
		  	 //se o handler ainda nao foi executado, executar
		  	 //comeca quando a tarefa de leitura terminar
			 {
			   wait_infos[c].any_fail=1;
			   if (wait_infos[c].existsTask!=0)
			     {														
				  DEBUG_MSG(TFD_LEVEL,"Machine %u Not Present, Starting failure handler <seq %u, to %u>\n",a,wait_infos[c].seq,wait_infos[c].to);		          
	  	 		  rt_task_resume(&(wait_infos[c].rt_task_id));
	  	 	     }	  	 		   
		 	 } 		  	 
		  	}
		  	else
		  	 {
		  	    //ainda podemos receber uma resposta	
		  	    if ((wait_infos[c].failure_info[a].fail==2)&&(system_view[a]==NORMAL))
		  	     can_delivery=0;
		  	 }
	  		} 
		  } //  for a machines
	  
		  if (can_delivery)		 { 
  		    reply_wInfo(c,wait_infos[c].send_fifo,wait_infos[c].destination_port);
		}  	    
  	    
		}  // time >
	} //for c
}

#endif

/*
+--------------------------------------------------------+
|  Constroi o conteudo dos pacotes enviados na rede TR.  |
|  Percorre a tabela send_info, que contem pedidos de    | 
|  detecao pendentes, e constroi os requests especificos.| 
|  Percorre tabela de recive_info: envia info de falhas  |  
+--------------------------------------------------------+ 
*/

#if defined (DIST_FAULT) || defined (DIST_AGREEMENT)
int buildDataInfo(msgtr_hdr * msg,unsigned long long ts)
{
	int items=0;
#ifdef DIST_FAULT
	int a;	
	long time_to_failure=0;
	unsigned long seq_to=0;
	unsigned long past;
#endif
	info_field* pt=(info_field*)(msg+1);	
#ifdef DIST_AGREEMENT
	CSTable_field *cst_ptr;
	int nfields_agr;
	int i;
	TCBtimestamp actTime;
#endif
		
#ifdef DIST_FAULT	
	//verifica pedidos de detecao pendentes 
	for (a=0;a<MAX_SEND_TABLE_SIZE;a++)
	{
		if (!send_info[a].is_free)
		{
			// quanto tempo passou desde o instante marcado como
			// o inicio da medicao e o evento que assinala 
			// a transmissao da mensagem (do mecanismo de rndtrip)?	
			past=ts-send_info[a].time_of_generation;
			past/=1000;  // em us
			time_to_failure=send_info[a].spec-past;
			// ainda nao falhou ?
			if (time_to_failure>=0)
			   time_to_failure=time_to_failure&0x3FFFFFFF;	
			 else // o tempo que falta para falhar e negativo
			  {
			   time_to_failure*=-1;
			   // 31 bits de time_to_failure em complemento 
			   // para dois e negativo
			   time_to_failure=time_to_failure|0x40000000;
			  }
			//Set request bit=0 - significa pedido
			time_to_failure=time_to_failure&0x7FFFFFFF;  
			seq_to=send_info[a].to;
			seq_to=seq_to<<24;
		        seq_to=send_info[a].seq|seq_to;
			items++; 
			send_info[a].is_free=1;
			pt->destination_port=send_info[a].destination_port;
			pt->r_time_to_failure=time_to_failure;
			pt->seq_to=seq_to; 
			pt++;									
			DEBUG_MSG(TFD_LEVEL,"Send TFD Invocation: to:<%u> from<%u> seq<%u>\n",send_info[a].to,tcb_machine_id,send_info[a].seq);
		}
	}
	
	//receive_info table
	for (a=0;a<MAX_RCV_TABLE_SIZE;a++)
	{
		if (!receive_info[a].is_free)
		{
			if (
			   // tempo expirado  	
			   ((ts>=receive_info[a].time_of_failure)&&(receive_info[a].fail==1))
			   ||
			   //nao falhou enviar a resposta
			   (receive_info[a].fail==0)
			   )
			{
			DEBUG_MSG(TFD_LEVEL,"Send TFD Reply: to:<%u> from<%u> seq<%u>\n",receive_info[a].from,tcb_machine_id,receive_info[a].seq);
		//printk("RESPOSTA\n");		
		//time_to_failure=rt_get_time_ns();
		//rt_printk("TOR: %X %X\n",(unsigned long)(time_to_failure>>32),(unsigned long)rt_get_time_ns());
		//time_to_failure=receive_info[a].time_of_failure;
		//rt_printk("TOR: %X %X\n",(unsigned long)(time_to_failure>>32),(unsigned long)receive_info[a].time_of_failure);
		
			if (receive_info[a].fail==0)
			{
			//rt_printk("NO FAIL\n");
				if (receive_info[a].lateness_available==1)
				{
				 time_to_failure=receive_info[a].lateness_degree;	 			 
	 			 //request bit =1  no fail bit=0 no lateness_degree bit=1
	 			 time_to_failure=time_to_failure&0x1FFFFFFF;
	 			 time_to_failure=time_to_failure|0xA0000000;
	 			} 
	 			 else
	 			  {
	 			   time_to_failure=0;	 			 
	 			   //request bit =1  no fail bit=0 no lateness_degree bit=0
	 			   time_to_failure=time_to_failure&0x1FFFFFFF;
	 			   time_to_failure=time_to_failure|0x80000000;
	 			  }
			}
			
			if (receive_info[a].fail==1)
			{
			//rt_printk("FAIL\n");
				if (receive_info[a].lateness_available==1)
				{
				 time_to_failure=receive_info[a].lateness_degree;	 			 
	 			 //request bit =1  fail bit=1 no lateness_degree bit=1
	 			 time_to_failure=time_to_failure|0xE0000000;
	 			} 
	 			 else
	 			  {
	 			   time_to_failure=0;	 			 
	 			   //request bit =1  fail bit=1 no lateness_degree bit=0
	 			   time_to_failure=time_to_failure&0x1FFFFFFF;
	 			   time_to_failure=time_to_failure|0xC0000000;
	 			  }
			}
			
		//	rt_printk("LD:%u\n",time_to_failure&0x1FFFFFFF);
			
		  	seq_to=receive_info[a].from;
		  	seq_to=seq_to<<24; 
		  	seq_to=receive_info[a].seq|seq_to;
		  	items++; 
		  	pt->r_time_to_failure=time_to_failure;
		  	 pt->seq_to=seq_to; 	
		  	pt->destination_port=receive_info[a].destination_port; 										  	
		  	// B bit
			if (receive_info[a].to!=my_mask)  //set B bit
			 	pt->seq_to=pt->seq_to|0x00800000;
			  else			      // clear B bit
			   	pt->seq_to=pt->seq_to&0xFF7FFFFF;
		  	receive_info[a].is_free=1;
		        pt++;
			}						
		}
	}		
	
	if (items<max_fields)
	{
		pt->seq_to=0x00000000;  //assinala final dos campos
		items++;
		pt++;	
	}
#endif //DIST_FAULT
	
#ifdef DIST_AGREEMENT
	cst_ptr = (CSTable_field *)pt;
	nfields_agr = first_CSTable_reg_free;

	if (nfields_agr > 0) 
	{
	  //rtl_printf("send nfields_agreement: %u\n", nfields_agr);
		memcpy(cst_ptr, CSTable, sizeof(CSTable_field)*nfields_agr);
		first_CSTable_reg_free = 0;
	}

	if (nfields_agr < max_fields_agreement)
	{
		cst_ptr = cst_ptr + nfields_agr;
		cst_ptr->tstart = 0;
		nfields_agr ++;
	}

	// discard old agreements data
	actTime=get_vclock();
	for (i = 0; i < MAX_AGREEMENTS; i++)
		if ((CDTable[i].is_free == 0) && (actTime > CDTable[i].tstart + TIME_DISCARD_AGREEMENT_RESULTS)) 
		CDTable[i].is_free = 1;


// TTCB -- deixou de retornar a var items, usada para calcular o tamanho do pacote e
// passou a retornar o tamanho do mesmo, excepto header
return (sizeof(info_field)*items + sizeof(CSTable_field)*nfields_agr);

#else //TTCB
	return items;
#endif	//TTCB
	
}

#endif //DIST_FAULT || DIST_AGREEMENT

/*
+----------------------------------------------------------------+
|  Actualiza time_to_failure (+diff) dos pedidos em send_info	 |
+----------------------------------------------------------------+ 
*/

#ifdef DIST_FAULT
void actualize_send_fields(msgtr_hdr * msg,unsigned long diff)
{
	int a;
	info_field* pt=(info_field*)(msg+1);
	long diff_in_micros=diff/1000;
		
	
	//send_info table
	for (a=0;a<MAX_SEND_TABLE_SIZE;a++)
	{
		if (!send_info[a].is_free)
		{
			pt->r_time_to_failure-=diff_in_micros;
			pt++;
		}	
	}	
}

#endif //DIST_FAULT


/*
+-----------------------------------------------------------+
|  Reinicia a informacao relativa a uma determinada maquina |
+-----------------------------------------------------------+ 
*/
#ifdef DISTRIBUTED
void reinit_infoOf(unsigned char id)
{
	int a;

#ifdef DIST_FAULT
	for (a=0;a<MAX_RCV_PLOAD_TABLE_SIZE;a++)
	{
	 if ((!receive_payload_info[a].is_free)&&(receive_payload_info[a].from==id))
	   receive_payload_info[a].is_free=1;
	} 

	for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++)
	{
	if ((!wait_infos[a].is_free)&&(wait_infos[a].from==id))
	   wait_infos[a].is_free=1;
	} 
	 
	for (a=0;a<MAX_RCV_TABLE_SIZE;a++)
	{
	if ((!receive_info[a].is_free)&&(receive_info[a].from==id))
	   receive_info[a].is_free=1;
	}
#endif //DIST_FAULT

//rt_printk("RINIT\n");
	system_view[id]=IN_INIT;
	gclock_info[id].vclock=0;
	gclock_info[id].error=0xFFFFFFFFFFFFFFF;
	number_of_active_machines=get_number_of_normal_state_machines();
}
#endif //DISTRIBUTED

/*
+--------------------------------------------------------------------------+
|              PROTOCOLO DE DESSIMINACAO PERIODICA E ATEMPADADA            |**
+--------------------------------------------------------------------------+**
   ***************************************************************************
*/


/*
+----------------------------------------------+
| ESTIMA O TEMPO DE ENTREGA DE UMA MENSAGEM    |
+----------------------------------------------+       
*/
#ifdef DISTRIBUTED
unsigned long check_round_trip(msgtr_hdr *msg)  
{
	unsigned long roundtripVal;
#ifdef  CONFIG_SELF_CHECKING_CONTROL_CHANNEL_REAL
	unsigned long irqstate;
#endif
 	
 	if (msg->RoundTripInfo[tcb_machine_id].rndval==0)
 	{
 	   closeReceive();
#ifdef RELEASE 
 	   while (1) {
#endif
 		rt_printk("TTCB PANIC: No round trip info in msg\n");
#ifdef RELEASE 	
 	   }
#endif
 	}
	
 	roundtripVal=(unsigned long)(last_rcv_info[msg->machine_src_id].timestamp-msg->RoundTripInfo[tcb_machine_id].snd_time)/1000;
  	roundtripVal=roundtripVal-msg->RoundTripInfo[tcb_machine_id].rndval;  //in microseconds
  	
   	// como o full-duplex e feito por software, nao se considera o valor
   	// de transmissao minimo quando a mensagem foi enviada e recebida pela
   	// mesma maquina.
  	if (msg->machine_src_id!=tcb_machine_id)  
  	 roundtripVal-=MINIMUM_DELIVERY_DELAY;	

#ifndef BENCH

	if (roundtripVal>MAXIMUM_ROUNDTRIP)
	{
#ifdef  CONFIG_SELF_CHECKING_CONTROL_CHANNEL_SIMULATED
		while (1){
#endif	
		 	rt_printk("TTCB Warning:  self checking for maximum message delivery delay activated - a value of %u us was estimated when the maximum allowed is %u \n",roundtripVal,MAXIMUM_ROUNDTRIP);
#ifdef  CONFIG_SELF_CHECKING_CONTROL_CHANNEL_SIMULATED	 
	 	};	
#endif
	} 

#ifdef  CONFIG_SELF_CHECKING_CONTROL_CHANNEL_REAL
	if (roundtripVal>MAXIMUM_ROUNDTRIP)
	{
	// shutdown 	
	irqstate = rt_global_save_flags_and_cli();
	machine_restart(&p);
	} 
#endif

#endif

#ifdef BENCH
	if (roundtripVal>max_roundtrip)
	  max_roundtrip=roundtripVal;	
#endif

 	return roundtripVal; 	
}



unsigned long check_round_trip_no_error(msgtr_hdr *msg)  
{
	unsigned long roundtripVal;
 	
 	if (msg->RoundTripInfo[tcb_machine_id].rndval==0)
 	{
 		return 0;
 	}
	
 	roundtripVal=(unsigned long)(last_rcv_info[msg->machine_src_id].timestamp-msg->RoundTripInfo[tcb_machine_id].snd_time)/1000;
  	roundtripVal=roundtripVal-msg->RoundTripInfo[tcb_machine_id].rndval;  //in microseconds
  	
   	// como o full-duplex e feito por software, nao se considera o valor
   	// de transmissao minimo quando a mensagem foi enviada e recebida pela
   	// mesma maquina.
  	if (msg->machine_src_id!=tcb_machine_id)  
  	 roundtripVal-=MINIMUM_DELIVERY_DELAY;	

 	return roundtripVal; 	
}

#endif //DISTRIBUTED


/*
+----------------------------------------------+
| REINICIA O PROTOCOLO DE INICIALIZACAO DO     |
| SISTEMA.				       |
+----------------------------------------------+       
*/

#ifdef DISTRIBUTED

extern void get_random_bytes(unsigned char*, unsigned char);

void reinit_protocol(msgtr_hdr *hdr)
{
	int a;
	unsigned char rndcont;

		get_random_bytes(&rndcont,1);   
		if (rndcont<10)
		 rndcont+=10;
//MUDAR PARA DEBUG_MSG	 
#ifdef DEBUGNET
		rt_printk("Reiniciando o protocolo - cont=%u\n",rndcont);
#endif 		   
		state_of_operation=INIT_STATE;
		contention_rounds=(unsigned char)rndcont;
		for (a=0;a<NUMBER_OF_MACHINES;a++)
		{
		 system_view[a]=NOT_PRESENT;
		 last_rcv_info[a].seq=0;
		 last_rcv_info[a].snd_time=0;
	 	 last_rcv_info[a].timestamp=0;
		 gclock_info[a].vclock=0;
	 	 gclock_info[a].error=0xFFFFFFFFFFFFFFF;
		}
		system_view[tcb_machine_id]=IN_INIT;
		for (a=0;a<NUMBER_OF_MACHINES;a++)
		  hdr->RoundTripInfo[a].rndval=0;
		number_of_active_machines=0;  
		wait_round=1;	
}
#endif

/*
+----------------------------------------------+
| INFORMA O DRIVER DA REDE DE CONTROLO DE QUE  |
| PODE COMECAR A RECEBER OS PACOTES.	       |
+----------------------------------------------+       
*/
#ifdef DISTRIBUTED
void openReceive(void)
{
	unsigned long irqstate;
	
	irqstate = rt_global_save_flags_and_cli();
   	startReceive();	 
	rt_global_restore_flags(irqstate);
}


/*
+----------------------------------------------+
| DIZ AO DRIVER DA REDE DE CONTROLO PARA       |
| IGNORAR OS PACOTES RECEBIDOS.		       |
+----------------------------------------------+       
*/

void closeReceive(void)
{
	unsigned long irqstate;
	
	irqstate = rt_global_save_flags_and_cli();
   	startReceive();	  
	rt_global_restore_flags(irqstate);
}
#endif //DISTRIBUTED


/*
+-----------------------------------------------------+
| ENVIA UMA MENSAGEM DE INICIALIZACAO DO PROTOCOLO    |
| DE DESSIMINACAO...NA REDE DE CONTROLO		      |
+-----------------------------------------------------+       
*/

#ifdef DISTRIBUTED
void send_init_message(void)  
{
	msgtr_hdr *hdrSnd=(msgtr_hdr *)&net_msg_memory;	
	
	hdrSnd->seq=1;	
	hdrSnd->msg_type=INIT_MSG;
	hdrSnd->snd_time=rt_get_time_ns();
	hdrSnd->vclock = 0;
//VER ISTO: next version
 	sendMessage(net_msg_memory,sizeof(msgtr_hdr),NETWORK_OMISSION_DEGREE+1); 	
 	hdrSnd->seq++;	
}

#endif //DISTRIBUTED

/*
+-----------------------------------------------------+
| DETECTOR DE PARAGENS				      |
+-----------------------------------------------------+       
*/

/*
  detectar falha (crash) quando
  
  Tempo mximo entre 2 recepcoes de Q from P (Ddecisao)
  
  
       c    Penvio    
P  ----|-------------|-----------------------
       | min          \ max
        |               \
Q  -----V--------------->---------|-----------
        a  	 	b 
        
  Ddecisao = Penvio - min + (ExecMaxenvio - ExecMinEnvio) + max      
 
  supondo que o processo para no instante c apos enviar uma msg, considera-se
  que a msg enviada demora o tempo maximo. Assim sendo, o crash e detectado 
  Ddecisao apos a recepcao da mensagem:		
  
         c    Penvio    
  P  ----|X------------|-----------------------
          | max          
           |               
  Q  -------V-------------------------------------
            y	 	       Q = y+Ddecisao
 
 	 
  Q detecta a falha de P no instante Q=c + max + Ddecisao
  
  Dcrash = Penvio(=SEND_PERIOD*(OMISSION_DEGREE-1))-min+(ExecMaxEnvio-ExecMinEnvio)+2*max+l	 
  
  l=margem de seguranca: devido aos desvios dos relogios e outros factores 
                         temos de dar uma margem de seguranca... 
  
  -----------------------------------
  PARA OPTIMIZAR NAS PROXIMAS VERSOES
  -----------------------------------
  Considera-se ExecMaxEnvio = 1ms e ExecMinEnvio = 0    	
  
  a vista e processada no inicio da tarefa de leitura, por isso, no pior caso:
 	 		 
  Dcrash + = 2*Pleitura + ExecMaxLeitura
  
  Considera-se ExecMax-Leitura = 1ms e ExecMinLeitura = 0    	
 
*/ 	 		 

#ifdef DISTRIBUTED
void process_system_view(void)
{

#ifdef DEBUGNET	
	char old_view[NUMBER_OF_MACHINES];  
#endif
	int a;

#ifdef DEBUGNET	
memcpy(&old_view,&system_view,NUMBER_OF_MACHINES);   
#endif
	
	number_of_active_machines=0;
	for (a=0;a<NUMBER_OF_MACHINES;a++)    //CRASH detector
	{
	if ((a!=tcb_machine_id)&&(system_view[a]==NORMAL))
	{
//NOTA: o valor 1ms esta hardcoded!!!
//VER BEM ISTO !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	 if (last_rcv_info[a].timestamp<(rt_get_time_ns()-(1000000+(SEND_PERIOD*(NETWORK_OMISSION_DEGREE+1))+SECURITY_MARGIN_FOR_NODE_CRASH_DETECTION+((2*MAXIMUM_ROUNDTRIP*1000)-(MINIMUM_DELIVERY_DELAY*1000))))){
	   system_view[a]=NOT_PRESENT;
	   gclock_info[a].vclock=0;
	   gclock_info[a].error=0xFFFFFFFFFFFFFFF;
	  }
	}
	if (system_view[a]==NORMAL)
	  number_of_active_machines++;
	}   

#ifdef DEBUGNET	
if (memcmp(old_view,system_view,NUMBER_OF_MACHINES)!=0)   //DEBUG
 {
 	for (a=0;a<5;a++)
 	 if (system_view[a]==NORMAL)
 	 {
 	  rt_printk("%u-",a);
 	 }
 	 rt_printk("\n");
 }
 #endif
		
}
#endif //DISTRIBUTED



/*
+-------------------------------+
| TAREFA DE LEITURA DO SISTEMA  |
+-------------------------------+       
*/
#ifdef DISTRIBUTED
void  ReadTask(int arg)
{
	int ret;
	RTIME t;
	msgtr_hdr *hdr=NULL;
	unsigned char rcv_msg_memory[MAX_MTU];
	unsigned char msgfrom_in_this_read_slot[NUMBER_OF_MACHINES];
	char msg_rcv;
	char printk_write=0;
#ifdef DEBUGNET
	int a;  
#endif
	unsigned long rndTrip;
	
#ifdef BENCH
	RTIME ti,te,tdiff;	
#endif	
		
	rt_task_make_periodic(rt_whoami(),nano2count(start_of_system_tasks+(SEND_PERIOD/2)),nano2count(READ_PERIOD));	

	openReceive();	   	
								
	while (1) {	
 	  rt_task_wait_period();
 	  printk_write=0;
 	  DEBUG_MSG(NETWORK_LEVEL,"Starting Read Task\n");
	  TCBenter_in_critical_section();
#ifdef BENCH
	ti=rt_get_time_ns();
	tdiff=ti-next_release_of_read;
	if (tdiff>max_release_time_read)
	 max_release_time_read=tdiff;
	next_release_of_read+=READ_PERIOD; 
#endif 	  
	memset(&msgfrom_in_this_read_slot,0,NUMBER_OF_MACHINES); 
 	
 	do {	
 rread_msg:
 		ret=readMessage((unsigned char*)&(rcv_msg_memory),&t);     
 		if (ret==0) goto no_msg;
 		hdr=(msgtr_hdr *)&(rcv_msg_memory); 	  	
 		if (hdr->magic_id!=0xF1){
 			if (!invalid_msg_display){
 				rt_printk("Warning: invalid packet on the channel....\n");
 				printk_write=1;
 				invalid_msg_display=1;
 			}
  			goto rread_msg;
 		}
 		msg_rcv=1;
 
 		if (hdr->machine_src_id>=NUMBER_OF_MACHINES)
 	  	 { 	  	  
 			rt_printk("Warning: invalid machine ID....\n");
 			printk_write=1;
  			goto rread_msg;
 	  	 }
 
	 	if (state_of_operation==SYNC_STATE)
 	  	{
		   state_of_operation=INIT_STATE;
		   read_loops=0;
		   send_init_message();
		   wait_round=1;
		} 
 
 	  	 	  
 	  	if ((state_of_operation==INIT_STATE)&&(contention_rounds>0))
 	  	 	goto rread_msg; 	  	 	  	
 
 		if (((hdr->msg_type==INIT_MSG))&&(state_of_operation!=NORMAL_STATE)&&(hdr->machine_src_id!=tcb_machine_id))  		 
 		{															
 		   //give up and reinit the protocol
#ifdef DEBUGNET 		   
		rt_printk("Recebi Init em init\n");
#endif		
 		   ret=1;
 		   //flush the message buffer
 		   while (ret!=0)
 		   	ret=readMessage((unsigned char*)&(rcv_msg_memory),&t);       
 		   reinit_protocol(hdr);		
 	   	   goto no_msg;
 		}
 	  		  	 
 	  	if ((hdr->msg_type==INIT_MSG)&&(hdr->machine_src_id!=tcb_machine_id)) // flush de seq info from the machine
 	  	 {
 	  	 	  last_rcv_info[hdr->machine_src_id].seq=0; 	  	 	  
 	  	 	  reinit_infoOf(hdr->machine_src_id);
#ifdef DEBUGNET
		rt_printk("Recebi init de %d\n",hdr->machine_src_id);
#endif 	  	 	  
 	  	 }  	  	  	  	 
		
		if (last_rcv_info[hdr->machine_src_id].seq==0)
		{
 	  	   	last_rcv_info[hdr->machine_src_id].seq=hdr->seq;
			last_rcv_info[hdr->machine_src_id].timestamp=t;   
			last_rcv_info[hdr->machine_src_id].snd_time=hdr->snd_time;
		}
		else
		{
 	  	
 		//duplicated!!!! 	  	  	  	
 	  	if (hdr->seq==last_rcv_info[hdr->machine_src_id].seq)
 	  	 {
		  goto rread_msg;
 	  	 }
 	  	  else
 	  	   if
 	  	   ( 
 	  	    (hdr->seq==last_rcv_info[hdr->machine_src_id].seq+1) ||
 	  	    ((hdr->seq==1)&&(last_rcv_info[hdr->machine_src_id].seq==255))
 	  	   ) 
 	  	   {
 	  	   	if (msgfrom_in_this_read_slot[hdr->machine_src_id])
 	  	   	 { 	  	   	   
 				rt_printk("Warning: two Messages from the same machine on the same read slot...\n");
 				printk_write=1;
 	  	   	 }
 	  	   	  else
 	  	   	   msgfrom_in_this_read_slot[hdr->machine_src_id]=1;
 	  	   	
 	  	   	last_rcv_info[hdr->machine_src_id].seq=hdr->seq;
			last_rcv_info[hdr->machine_src_id].timestamp=t;   
			last_rcv_info[hdr->machine_src_id].snd_time=hdr->snd_time;
	 	  }	
	      	   else
	      	          {					 	            
				rt_printk("Warning: message lost from %u...\n",hdr->machine_src_id);	
				printk_write=1;
				last_rcv_info[hdr->machine_src_id].seq=hdr->seq;
				last_rcv_info[hdr->machine_src_id].timestamp=t;   
				last_rcv_info[hdr->machine_src_id].snd_time=hdr->snd_time;
	 	          }  	 	    	 		        
	         }  //seq!=0

	goto rcvd_msg;	
no_msg:
	msg_rcv=0;
	if (read_loops==SYNC_ROUNDS)
	  {
	    wait_round=1;      // I'm the only 
	    state_of_operation=INIT_STATE;		
	    send_init_message();	
	    read_loops=0;
#ifdef DEBUGNET
	  rt_printk("No activity on SYNC\n");	
#endif	    
	  }		
rcvd_msg: 
	if ((state_of_operation==INIT_STATE)&&(contention_rounds>0))
 	   {
	    contention_rounds--;
	    if (contention_rounds==0)
	     send_init_message();
	   }

	  if (wait_round==0)
	   read_loops++;
	 	 	   
	//TODO ver implicacoes!!!!!!!!!!!!!!!!!!!!
 	  if (state_of_operation==NORMAL_STATE)	 
   	    process_system_view();
	 
 	 if (msg_rcv){
 
  	     DEBUG_MSG(NETWORK_LEVEL,"Message From <%u>, seq <%u>\n",hdr->machine_src_id,hdr->seq);
 	     
 	     if ((state_of_operation==INIT_STATE)&&(contention_rounds==0)){
 	    	system_view[hdr->machine_src_id]=NORMAL;   
		rndTrip=check_round_trip_no_error(hdr);
		update_vclock(hdr->machine_src_id, hdr->vclock, rndTrip, last_rcv_info[hdr->machine_src_id].timestamp);
 	     }
 	     else
 	      if ((state_of_operation==NORMAL_STATE)&&(hdr->msg_type!=INIT_MSG))
 	      {  	         	      	   
 	        if (system_view[hdr->machine_src_id]!=NOT_PRESENT)
 	        {   
 		  rndTrip=check_round_trip(hdr);
 
 	          if (system_view[hdr->machine_src_id]==IN_INIT)
 	             {
 	              system_view[hdr->machine_src_id]=NORMAL;
 	              number_of_active_machines++;
#ifdef DEBUGNET 	              
 	              rt_printk("NN:%u\n",number_of_active_machines);
#endif
 	             } 
   		     else
   	         	update_vclock(hdr->machine_src_id, hdr->vclock,rndTrip, last_rcv_info[hdr->machine_src_id].timestamp);

#if defined(DIST_FAULT) || defined(DIST_AGREEMENT)   
 	          process_read(hdr,rndTrip,last_rcv_info[hdr->machine_src_id].timestamp);
#endif
 
 	         }
 	          else
 	           {
 	         	   	rt_printk("Warning: message received from a machine not present in the system view...\n" );	
 				printk_write=1;
 				system_view[hdr->machine_src_id]=NORMAL_STATE;
 				number_of_active_machines++;
 				
 	           } 
 	      }     
 	     
	   }//msg_rcv	

#ifdef DIST_FAULT 	   
 	   if (state_of_operation==NORMAL_STATE)
 	   {
 	     //para o caso de ocorrerem paragens no sistema	
	     process_wInfos();		
 	   }
#endif
 	    	    	   
 	   if ((state_of_operation==INIT_STATE)&&(wait_round==10))
 	   {
 	     rt_printk("\n *** TTCB has been started and is running  ***\n\n");	
 	     printk_write=1;
 	     state_of_operation=NORMAL_STATE;
 	     system_view[tcb_machine_id]=NORMAL;
 	     init_vclock();	
 	     wait_round=1;
 	     number_of_active_machines=get_number_of_normal_state_machines();
#ifdef DEBUGNET
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	 if (system_view[a]==NORMAL)
	  rt_printk("%d-",a);
	rt_printk("\nNM:%u\n",number_of_active_machines);   	  	
#endif 	      
 	   }
 	   
 	   if ((state_of_operation==INIT_STATE)&&(contention_rounds==0)) wait_round++;
 	   	
 	} while (msg_rcv==1);				
 
 	   
#ifdef BENCH
	if (printk_write==0){
		te=rt_get_time_ns();
		tdiff=te-ti;
		if (tdiff>bench_max_exec_read)
		  bench_max_exec_read=tdiff;
	}
#endif 	 
	TCBleave_critical_section();
	DEBUG_MSG(NETWORK_LEVEL,"Ending Read Task\n");
	}//end while(1)
}
#endif  //DISTRIBUTED


/*
+------------------------------------------------------+
| COLOCA NO PACOTE A INFORMACAO RESPECTIVA AO MECANISMO|
| DE CALCULO DE TEMPOS DE IDA-E-VOLTA DAS MENSAGENS.   |
| DEVOLVE O INSTANTE ASSINALADO PARA O INICIO DA       | 
| TRANSMISSAO.				               |            
+------------------------------------------------------+       
*/

#ifdef DISTRIBUTED
unsigned long long buildRoundTripInfo(msgtr_hdr * msg)
{
	int a;
	unsigned short valor;
	RTIME now=rt_get_time_ns();
	
   	for (a=0;a<NUMBER_OF_MACHINES;a++)
   	if (system_view[a]!=NOT_PRESENT)
   	     {
   	      valor=(unsigned long)(now-last_rcv_info[a].timestamp)/1000;   	      
   	      msg->RoundTripInfo[a].rndval=valor;
   	      msg->RoundTripInfo[a].snd_time=last_rcv_info[a].snd_time;
   	      msg->snd_time=now;
   	     } 
   	return now;  
}
#endif

#ifdef DISTRIBUTED
/*
+-----------------------------+
| TAREFA DE ENVIO DO SISTEMA  |
+-----------------------------+       
*/

void  SendTask(int arg)
{
	msgtr_hdr *hdr=(msgtr_hdr *)&net_msg_memory;
	unsigned long irqstate;
	unsigned long long ts,ts2;
	int n_fields=0;
#ifdef BENCH
	RTIME ti,te,tdiff;
#endif	

	start_of_system_tasks=rt_get_time_ns();
	
#ifdef BENCH	
	next_release_of_send=start_of_system_tasks+SEND_PERIOD;
	next_release_of_read=start_of_system_tasks+(SEND_PERIOD/2)+READ_PERIOD;
#endif
   
	rt_task_make_periodic(rt_whoami(),nano2count(start_of_system_tasks),nano2count(SEND_PERIOD));   		
   
	while (1)
	{
 	  rt_task_wait_period();
 	  DEBUG_MSG(NETWORK_LEVEL,"Starting Send Task\n");
 	  TCBenter_in_critical_section();  
 
#ifdef BENCH
	ti=rt_get_time_ns();
	tdiff=ti-next_release_of_send;
	if (tdiff>max_release_time_send)
	   max_release_time_send=tdiff;
	  next_release_of_send+=SEND_PERIOD; 
#endif	 
	
	   if (state_of_operation==NORMAL_STATE) 
	    {
	     hdr->msg_type=NORMAL_MSG;
	     if (od_counter == 0){	
		     ts=rt_get_time_ns();
#if defined (DIST_FAULT) || defined (DIST_AGREEMENT)	     
		     n_fields=buildDataInfo(hdr,ts);  	     
#else
		     n_fields=0;		     
#endif	     
		     irqstate = rt_global_save_flags_and_cli();
		     ts2=buildRoundTripInfo(hdr);
		     //to provide a better accuracy on the FD		 	     
#ifdef DIST_FAULT	     
		     actualize_send_fields(hdr,ts2-ts);	 	
#endif	     
		     hdr->vclock = get_vclock_at(ts2); 	
	     	     send_time = ts2;	
	     }	
	     else {
		     irqstate = rt_global_save_flags_and_cli();	
		     ts2=buildRoundTripInfo(hdr);
#ifdef DIST_FAULT	     	     
		     actualize_send_fields(hdr,ts2-send_time);	 	
#endif	     
		     hdr->vclock = get_vclock_at(ts2); 	
	     }    
//PARA MUDAR....
#ifdef DIST_AGREEMENT
	   sendMessage(net_msg_memory,sizeof(msgtr_hdr)+(n_fields),1);     
#else
 	   sendMessage(net_msg_memory,sizeof(msgtr_hdr)+(sizeof(info_field)*n_fields),1);     
#endif
 	   rt_global_restore_flags(irqstate); 	  	    	    
 
 	   if (od_counter == 0){	
	     hdr->seq++;	
	     if (hdr->seq==0)
	      hdr->seq++;
	     n_sends_in_this_slot=0; 
	   }	
	   od_counter++;	
	   if (od_counter==NETWORK_OMISSION_DEGREE+1){
		od_counter=0;
	   }	
	  }  

#ifdef BENCH
	te=rt_get_time_ns();
	tdiff=te-ti;
	if (tdiff>bench_max_exec_send)
	  bench_max_exec_send=tdiff;
#endif		
	TCBleave_critical_section();
	DEBUG_MSG(NETWORK_LEVEL,"Ending Send Task\n");
	}
}
#endif

/*
+-------------------------------------+
| Cria as tarefas de envio e leitura  |
+-------------------------------------+       
*/

#ifdef DISTRIBUTED
int createSendReceiveTask(void)
{	
	int ret;
	

	printk("\n-creating send task ");

					
	ret=rt_task_init(&snd_task_id, SendTask, 0, 10000, SYSTEM_TASK_PRIORITY, 1, 0);

	if (ret!=0)																			
	  {
	    printk("->failed\n");  
	    return -1;
          }
               
        rt_task_resume(&snd_task_id);
          
        printk("->succeded\n");
        
        printk("\n-creating read task ");
        
        
	ret=rt_task_init(&rcv_task_id, ReadTask, 0, 10000, SYSTEM_TASK_PRIORITY, 1, 0);
	
	if (ret!=0)																			
	  {
	    printk("->failed\n");  	
	    return -1;
          }
         
        printk("->succeded\n");    
          
        printk("\n*) waiting for the conclusion of the initialization protocol...\n");  
          
        rt_task_resume(&rcv_task_id);           
        
	return 0;
}

#endif //DISTRIBUTED

/*
+-----------------------+
| Inicia a comunicacao  |
+-----------------------+       
*/

#ifdef DISTRIBUTED
void init_networking(void)
{
	int a;
	msgtr_hdr *hdr=(msgtr_hdr *)&net_msg_memory;
	
	state_of_operation=SYNC_STATE;
	hdr->machine_src_id=get_tcb_id();
	hdr->magic_id=0xF1;
	tcb_machine_id=hdr->machine_src_id;
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	{
	 system_view[a]=NOT_PRESENT;
	 last_rcv_info[a].seq=0;
	 last_rcv_info[a].snd_time=0;
	 last_rcv_info[a].timestamp=0;
	}
	system_view[tcb_machine_id]=IN_INIT;
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	  hdr->RoundTripInfo[a].rndval=0;
	wait_round=0;
	read_loops=0;
	contention_rounds=0;
	printk("\n-TCB ID <%d>\n",hdr->machine_src_id);
	my_mask=1<<(tcb_machine_id);
	number_of_active_machines=0;
	hdr->seq=1;	
}
#endif

/*
+--------------------------------------------------------------------------+
| Devolve o numero de registos disponibilzados em cada mensagem ao servico |
| de deteccao de falhas distribuidas					   |
+--------------------------------------------------------------------------+       
*/
#if defined(DIST_AGREEMENT) || defined(DIST_FAULT)
void get_max_fields(void)
{
	int max_mtu;
	int net_mtu=getNetMTU();
	
	netMTU=net_mtu;
	
	max_mtu=NET_FIFO_SIZE/NUMBER_OF_MACHINES;
	
	if (max_mtu>net_mtu)
	 max_mtu=net_mtu;

//COMENTEI!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//	max_mtu=max_mtu/(NETWORK_OMISSION_DEGREE+1);	

#if defined(DIST_AGREEMENT) && defined(DIST_FAULT)
	max_mtu = max_mtu / 2;  //half to header+FD, another half to timestamp+Agreement
	max_fields_agreement = (max_mtu / sizeof(CSTable_field));
	DEBUG_MSG(NETWORK_LEVEL,"max_fields_agreement %d\n", max_fields_agreement);
#endif

#if defined(DIST_AGREEMENT) && !defined(DIST_FAULT)
	max_fields_agreement = ((max_mtu-sizeof(msgtr_hdr)) / sizeof(CSTable_field));
	DEBUG_MSG(NETWORK_LEVEL,"max_fields_agreement %d\n", max_fields_agreement);
#endif

#ifdef DIST_FAULT
	max_fields=(max_mtu-sizeof(msgtr_hdr))/sizeof(info_field);
	DEBUG_MSG(NETWORK_LEVEL,"max_fields_tfd %d\n", max_fields);		
	max_items=max_fields/(NUMBER_OF_MACHINES+1);		
	DEBUG_MSG(NETWORK_LEVEL,"max_items_tfd %d\n", max_items);		
#endif
}
#endif //DIST_AGREEMENT || DIST_FAULT


int init_module(void)
{			

 	// inicializa as variaveis do sistema
	
	initVariables();	
	
	printk("\nstarting ...\n\n");
	printk("+--------------------------+\n");
       	printk("| TTCB 1.1 for RTAI Linux  |\n");
#ifdef DEBUG_VERSION
	printk("| WARNING: DEBUG VERSION   |\n");
#endif
	printk("+--------------------------+\n");

	rt_set_oneshot_mode();
	rt_preempt_always(1);
	start_rt_timer(0);	

	// criar os fifos	
	
	if (createFifos())
	 return -1;
	 
	// associar um handler que e invocado quando chegam pedidos a tcb,
	// ou seja quando chega informacao ao fifo 
	
	if (setconHandler())
	 return -1;
	
#ifdef DISTRIBUTED 
	init_networking(); 
#endif

#if defined(DIST_AGREEMENT) || defined(DIST_FAULT)	
	get_max_fields(); 
#endif
	 
#ifdef DISTRIBUTED
	if (createSendReceiveTask())
	  return -1;
#endif 

#ifdef TTCB
	DEBUG_MSG(SECURITY_LEVEL,"\nGenerating secret key of size %u\n",1024);
		generate(&SecretKey,(unsigned)1024);

		if(check_secret_key(&SecretKey)==1){
			DEBUG_MSG(SECURITY_LEVEL,"Key checked OK!\n");
		}
		else 
		{
			DEBUG_MSG(SECURITY_LEVEL,"\n\n ***** Key failed!!! *****\n");
			return -1;
		}

		PublicKey.n=SecretKey.n;
		PublicKey.e=SecretKey.e;
#endif

#ifndef DISTRIBUTED
	printk("\n*) waiting for the conclusion of the initialization protocol...\n");
	printk("\n *** TTCB has been started and is running  ***\n\n");
#endif
		
	//kernel info structure - returned by get_kernel_info (T)TCB sys call
	kern_info.version = TCB_VERSION_CODE >>16;
	kern_info.sub_version = (TCB_VERSION_CODE >>8)&(0x00FF);
#ifdef TTCB
	kern_info.ttcb_extentions = 1;
#else
	kern_info.ttcb_extentions = 0;
#endif			
	return 0;
}

/*
+-------------------------------------+
| elimina tarefas pendentes           | 
+-------------------------------------+ 
*/

void killPendingTasks(void)  
{
	int a;
	unsigned long irqstate;
	 
	a=0;
	irqstate = rt_global_save_flags_and_cli();
		 	   
#ifdef LOCAL_TIMELY
	for (a=0;a<MAX_EXEC;a++)		
	{
		if (executions[a].func!=0)
		{
		   rt_task_delete(&(executions[a].task_id));		   
		}
		
	}
#endif //LOCAL_TIMELY

#ifdef LOCAL_FAULT
	    	
	for (a=0;a<MAX_LOCAL_FD;a++)		
	{
	if (localFD[a].func!=0)
		{
		   rt_task_delete(&(localFD[a].task_id));
		}
	}
	
#endif

#ifdef DIST_FAULT
	for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++)
	{
	if ((!wait_infos[a].is_free)&&(wait_infos[a].existsTask!=0))
		{
		   rt_task_delete(&(wait_infos[a].rt_task_id));
		}
	}
	
#endif

#ifdef DISTRIBUTED
	stopReceive();	  
	
        rt_task_delete(&snd_task_id);
        rt_task_delete(&rcv_task_id);
#endif //DISTRIBUTED
	rt_global_restore_flags(irqstate);	
}


void cleanup_module(void)
{	
		
	killPendingTasks();
	destroyFifos();
	printk("\n\nTTCB execution has been terminated\n\n");	

	stop_rt_timer();
}


/* From here on, it's just TTCB specific code
 */

#ifdef TTCB

#define HMAC_OPAD       0x5C
#define HMAC_IPAD       0x36



/*
+---------------------------+
| auxiliary functions       +
+---------------------------+
*/

// applies the private key to an input, cyphering or decyphering it
int TTCBprivate(void *input, void*output)
{

  MPI data;
  MPI result=mpi_alloc(1);

  data = mpi_alloc(sizeof(input)/sizeof(mpi_limb_t));
  data->nlimbs = sizeof(input)/sizeof(mpi_limb_t);
  data->d = input;

  secret(result, data, &SecretKey);

  output = result->d;

  return 1;

}


#ifdef DIST_AGREEMENT
// search for an agreement in CDTable given a unique agreement id: (elist, tstart, decision)
short search_CDTable(elist_t elist, TCBtimestamp tstart, decision_t decision)
{
        short i, j, size_el;

        for (i=0; i<MAX_AGREEMENTS; i++) {
                for(j=0; (j<MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST); j++)
                        ;
                size_el = j;
                if (j<MAX_SIZE_ELIST) size_el++; //compare also END_ELIST

                if ((CDTable[i].tstart == tstart) && (CDTable[i].decision == decision) &&
                        (memcmp(CDTable[i].elist, elist, size_el*sizeof(eid_t)) == 0) && (CDTable[i].is_free == 0))
                return i;
        }
        return -1;
}

#endif



#ifdef DIST_AGREEMENT
// search for a agreement in CDTable given the tag
short search_CDTable_tag(tag_t tag)
{
	short i;

	for (i=0; i<MAX_AGREEMENTS; i++)
		if ((CDTable[i].tag == tag) && (CDTable[i].is_free == 0))
			return i;
	return -1;
}

#endif

#ifdef DIST_AGREEMENT
// get a free register in CDTable
short get_freeR_in_CDTable(void)
{
	short i, j;
	

	DEBUG_MSG(INTERNAL_LEVEL,"Going to get_freeR_in_CDTable.\n");
	TCBenter_in_critical_section();
	for (i = 0; i<MAX_AGREEMENTS; i++)
	  if (CDTable[i].is_free == 1) {
	    
	    CDTable[i].is_free = 0;
	    for (j=0; j<MAX_SIZE_ELIST; j++) CDTable[i].vtable[j].is_free = 1;
	    TCBleave_critical_section();
	    DEBUG_MSG(INTERNAL_LEVEL,"Found a free Record at %d\n",i);
	    return i;
	  }
	
	TCBleave_critical_section();
	DEBUG_MSG(INTERNAL_LEVEL,"Did not find a free Record.\n");
	return -1;
}

#endif

#ifdef DIST_AGREEMENT
// count free registers in CDTable
short count_freeR_in_CDTable(void)
{
        short i, j;

	j = 0;
        for (i = 0; i<MAX_AGREEMENTS; i++)
                if (CDTable[i].is_free == 1) 
			j++;
        return j;
}

#endif

#ifdef TTCB
// get a free register in the table entities_data
short get_freeR_in_entities_data(void)
{
	short i;

        DEBUG_MSG(INTERNAL_LEVEL,"Going to get freeR in entities\n");
	TCBenter_in_critical_section();
	for (i = 0; i< MAX_LOCAL_ENTITIES; i++){
	  
	  if (entities_data[i].is_free) {
	    entities_data[i].is_free = 0;
	    entities_data[i].seq = 1;
	    TCBleave_critical_section();
	    DEBUG_MSG(INTERNAL_LEVEL,"Found a free entity at %i.\n",i);
	    return i;
	  }
	}
	
	TCBleave_critical_section();
	DEBUG_MSG(INTERNAL_LEVEL,"Did not find a free entity\n");
	return -1;
}
#endif

#ifdef DIST_AGREEMENT
// search for an eid in elist
short index_eid_in_elist(eid_t eid, elist_t elist)
{
	short i;

	for (i = 0; (elist[i] != END_ELIST) && (i< MAX_SIZE_ELIST); i++)
		if (elist[i] == eid)
			return i;
	return -1;
}
#endif




#ifdef DIST_AGREEMENT
// return a free tag
tag_t get_tag(void)
{
	last_tag = last_tag + 1;
	// here we should check if this one is already being used...
	return last_tag;
}
#endif




/*
+------------------------------------+
| inits TTCB variables               | 
+------------------------------------+ 
*/

void initVariablesTTCB(void)
{
  int a;
#ifdef DIST_AGREEMENT
  int b;
#endif

	for (a=0;a< MAX_LOCAL_ENTITIES;a++)
		entities_data[a].is_free=1;

#ifdef DIST_AGREEMENT
	for (a=0;a< MAX_AGREEMENTS;a++) {
		CDTable[a].is_free=1;
		for (b=0;b< MAX_SIZE_ELIST;b++)
			CDTable[a].vtable[b].is_free = 1;
	}

	first_CSTable_reg_free = 0;
	last_tag = 0;
#endif
	this_local_TTCB_id = get_tcb_id();
}



/*
+------------------------------------+
| input processing                   | 
+------------------------------------+ 
*/

// function that checks the secret or uncompresses data, and validates inputs
// 	reqHead has two fields: srvc (service requested) and sizeofmsg (size of in_)

void process_input(int fifo, ReqTcbHeader reqHead, in_hdr_t *in, int size_out)
{
	unsigned char out_buf[MAX_SIZE_OUT_STRUCT]; // enought space to hold any out struct, since this function does not know the precise type and may have to return an error
	error_out_t *out = (error_out_t *)out_buf;
	unsigned char Rindex;
	unsigned char mac[sizeof(skey_t)];

	Rindex = EIDtoINDEX(in->eid);
	if ((Rindex > MAX_LOCAL_ENTITIES) || (entities_data[Rindex].is_free == 1) || (EIDtoLTID(in->eid) != this_local_TTCB_id)) {
		out->error = INVALID_ID;
		process_output(fifo, (out_hdr_t *)out, size_out, in->eid, PROTECTION_NONE, in->secret_mic, in->seq);
		return;
	}
	
	switch(entities_data[Rindex].protection) {
	case PROTECTION_NONE:
	  DEBUG_MSG(INTERNAL_LEVEL,"Going to process input with no protection.\n");
	  if ( (memcmp(in->secret_mic, entities_data[Rindex].secret, sizeof(skey_t)) != 0) || (in->eid != in->copy_eid) ) {
	    out->error = INVALID_ID;
	    
	    process_output(fifo, (out_hdr_t *)out, size_out, in->eid, PROTECTION_NONE, in->secret_mic, in->seq);
	    return;
	  }
	  break;
	  
	case PROTECTION_HMD5:
	  DEBUG_MSG(INTERNAL_LEVEL,"Going to process input with MD5 protection.\n");
	  if ( in->copy_eid != in->eid ) {
	    out->error = INVALID_ID;
	    process_output(fifo, (out_hdr_t *)out, size_out, in->eid, PROTECTION_HMD5, in->secret_mic, in->seq);
	    return;
	  }
	  if (in->seq != entities_data[Rindex].seq) {

	    out->error = INVALID_SEQ;
	    process_output(fifo, (out_hdr_t *)out, size_out, in->eid, PROTECTION_HMD5, in->secret_mic, in->seq);
	    return;
	  }
	  entities_data[Rindex].seq ++;
	  
	  hash_md5((unsigned char *)&(in->copy_eid), reqHead.sizeofmsg -sizeof(eid_t)*2 - sizeof(skey_t), (unsigned char *)entities_data[Rindex].secret, sizeof(skey_t), mac);
	  if (memcmp(in->secret_mic, mac, sizeof(skey_t)) != 0) {
	    out->error = INVALID_ID;
	    process_output(fifo, (out_hdr_t *)out, size_out, in->eid, PROTECTION_HMD5, in->secret_mic, in->seq);
	    return;
	  }
	  break;
	  
	default:
	  DEBUG_MSG(SERVICE_LEVEL,"Fatal error in process_input\n"); // cannot happen
	}
	DEBUG_MSG(INTERNAL_LEVEL,"Finished processing input.\n");
	return;
}


/*
+------------------------------------+
| output processing                  | 
+------------------------------------+ 
*/

// function that protect output, sends it, etc
void process_output(int fifo, out_hdr_t *out, int len, eid_t eid, protection_t protection, skey_t secret, seq_t seq)
{

	DEBUG_MSG(INTERNAL_LEVEL,"Going to process output. Protection is %d\n",protection);

	out->eid = eid;
	out->seq = seq;
	

	DEBUG_MSG(INTERNAL_LEVEL,"Going to check protection type.\n");
	switch(protection) {
		case PROTECTION_NONE:	
 		        DEBUG_MSG(INTERNAL_LEVEL,"No protection\n");
			memcpy(out->secret_mic, secret, sizeof(skey_t));
			break;

		case PROTECTION_HMD5:
		        DEBUG_MSG(INTERNAL_LEVEL,"Protected by MD5\n");
			hash_md5((unsigned char *)&(out->eid), len - sizeof(skey_t), secret, sizeof(skey_t), out->secret_mic);
			break;
			
		default: 
			DEBUG_MSG(SERVICE_LEVEL,"Fatal error in process_output\n"); // cannot happen
        }

	DEBUG_MSG(INTERNAL_LEVEL,"Going to send output!\n");
	rtf_put(fifo, out, len);

	DEBUG_MSG(INTERNAL_LEVEL,"Output sent.\n");
	return;
}



/*
+-------------------+
|  TTCB SERVICES    |**********************************************************
+-------------------+
*/


/*
+----------------------+
| local authentication |
+----------------------+
*/

void TTCBlocalAuthentication(int fifo, char *buff, int sizeofmsg)
{
        LA_in_t in;
	LA_out_t out;
	int Rindex;
	MPI data;
	MPI result;

	MPI Ciphered;
	MPI deCiphered;
	char * temppointer;

	char secretbuffer[17]="\0";

	// local authent is an exception: output not handled by process_output

		  /* Decipher the input. It should come signed with our public key, so use our secret key */
	  Ciphered=mpi_alloc(sizeofmsg-24);
	  deCiphered=mpi_alloc(sizeof(in));

	  temppointer = (char *)Ciphered->d;

	  memcpy(Ciphered, buff, 24);
	  Ciphered->d = (mpi_limb_t *)temppointer;
	  memcpy(Ciphered->d, (buff + 24), (sizeofmsg - 24 ));

	  secret(deCiphered, Ciphered, &SecretKey); 

	  memcpy(&in, deCiphered->d, sizeof(in));


	  /* We now have the LA_in_t struct deciphered. */

	// VERIFY IF IN.PROTECTION IS MEANINGFUL/VALID
	// IF IT IS NOT, MAYBE THE DECRYPTION WENT WRONG

	if ((in.protection != PROTECTION_NONE) && (in.protection != PROTECTION_HMD5)) {
	  out.error = INVALID_PROTECTION;
	  rtf_put(fifo, &out, sizeof(out));
	  return;
	}

	if ((Rindex = get_freeR_in_entities_data()) == -1) {
		out.error = NO_ROOM_FOR_ENTITY;
		rtf_put(fifo, &out, sizeof(out));
		return;
	}

	DEBUG_MSG(INTERNAL_LEVEL,"Attributed Rindex %i.\n",Rindex);
	memcpy(entities_data[Rindex].secret, in.key, sizeof(skey_t));
	entities_data[Rindex].protection = in.protection;
	entities_data[Rindex].is_free = 0;
	out.error = 0;
	out.eid = toEID(this_local_TTCB_id, Rindex);
	// out.challenge_sign = secret( RESULT, IN->challenge, 16);

	memcpy(secretbuffer,entities_data[Rindex].secret,16);

	/* Going to cipher in.challenge */

	data=mpi_alloc(8);
	result= mpi_alloc(16);

	data->nlimbs = sizeof(challenge_t)/sizeof(mpi_limb_t);
	memcpy(data->d, in.challenge, 8);

	secret(result, data, &SecretKey);

	memset(out.challenge_sign,0,152);

	memcpy(out.challenge_sign, result, 24);
	memcpy((out.challenge_sign+24), result->d, (mpi_get_nbits(result)+7)/8);
  
	rtf_put(fifo, &out, 192); 
	return;
}


/*
+---------------------+
| end connection      |
+---------------------+
*/

void TTCBend_connection(int fifo, eid_in_t *in)
{
	error_out_t out;
        int Rindex;

        out.error = 0;

	DEBUG_MSG(SERVICE_LEVEL,"Going to end connection on fifo %i\n",fifo);
        Rindex = EIDtoINDEX(in->eid);
	entities_data[Rindex].is_free = 1;

	free_resources(fifo);
        usedFifos[fifo-TCB_CON_STARTFIFO]=false;

	process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);

        return; 
}



/*
+-------------+
| agreement   |
+-------------+
*/

#ifdef DIST_AGREEMENT
void TTCBpropose(int fifo, propose_in_t *in)
{
	propose_out_t out;
	TCBtimestamp actTime;
	short Rindex, // entity index in entities_data
	      Dindex, // agreement index in CDTable
	      index_in_elist; // index of the eid in elist
	tag_t tag=0;
	int j;

	out.error = 0;

	Rindex = EIDtoINDEX(in->eid);
	DEBUG_MSG(SERVICE_LEVEL,"Entering propose!\n");
	// tests
	if ((index_in_elist = index_eid_in_elist(in->eid, in->elist)) == -1) { // if eid not in elist then error
	        DEBUG_MSG(SERVICE_LEVEL,"EID_NOT_IN_ELIST\n");
	        out.error = EID_NOT_IN_ELIST;
	        process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
		return;
	}

	if (in->decision > MAX_NUMBER_DECISIONS) { // invalid decision
		out.error = INVALID_DECISION;
		DEBUG_MSG(SERVICE_LEVEL,"INVALID_DECISION\n");
		process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
		return;
	}

	if ((Dindex = search_CDTable(in->elist, in->tstart, in->decision)) != -1) {
	        DEBUG_MSG(INTERNAL_LEVEL,"Search_CDTable returned %d\n",Dindex);
		if (CDTable[Dindex].vtable[index_in_elist].is_free == 0) {
			out.error = VALUE_ALREADY_PROPOSED; // sorry, you already proposed
			DEBUG_MSG(SERVICE_LEVEL,"VALUE_ALREADY_PROPOSED\n");
			process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
			return;
		}
		tag = CDTable[Dindex].tag;
		DEBUG_MSG(INTERNAL_LEVEL,"The tag is %d\n",tag);
	}
	else  // there is still no such agreement: let's test if there's room for it
	  {
	    DEBUG_MSG(INTERNAL_LEVEL,"There is still no such agreement [%i]\n",in->eid);
	    if ((first_CSTable_reg_free == max_fields_agreement) ||
            	(count_freeR_in_CDTable() < max_fields_agreement * number_of_active_machines)) { // no space for propose
	      out.error = NO_RESOURCES_PROPOSE;
	      DEBUG_MSG(SERVICE_LEVEL,"NO_RESOURCES_PROPOSE\n");
	      process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
	      return;
	    }
	  }

	DEBUG_MSG(INTERNAL_LEVEL,"Entering critical section in propose.\n");
	TCBenter_in_critical_section();

	actTime=get_vclock();
	if (actTime > in->tstart) { // if tstart passed, error

		if (actTime >= in->tstart + TIME_DISCARD_AGREEMENT_RESULTS) 
			out.error = AGREEMENT_EXPIRED;
		else
		if (Dindex != -1) {  // if agreement exists lets return tag
			out.tag = CDTable[Dindex].tag;
			out.error = TSTART_EXPIRED;
		} else { // else let us create and return a tag
			tag = get_tag();
			if ((Dindex = get_freeR_in_CDTable()) == -1) {
				out.error = MAX_NUM_AGREEMENTS_RUNNING;
				process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
				return;
			}
			CDTable[Dindex].tag = tag;
			memcpy(CDTable[Dindex].elist, in->elist, sizeof(elist_t));
			CDTable[Dindex].tstart = in->tstart;
			CDTable[Dindex].decision = in->decision;
			out.tag = tag;
			out.error = TSTART_EXPIRED;
		}
		TCBleave_critical_section();
		process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
		DEBUG_MSG(SERVICE_LEVEL,"TSTART_EXPIRED\n");
		return;
	}

	memcpy(CSTable[first_CSTable_reg_free].elist, in->elist, sizeof(elist_t));
	CSTable[first_CSTable_reg_free].tstart = in->tstart;
	CSTable[first_CSTable_reg_free].decision = in->decision;
	CSTable[first_CSTable_reg_free].eid = in->eid;
	memcpy(CSTable[first_CSTable_reg_free].value, in->value, sizeof(value_t));

	first_CSTable_reg_free++;

	TCBleave_critical_section();

	DEBUG_MSG(INTERNAL_LEVEL,"Leaving critical section in propose...\n");

	if (Dindex == -1) {
		tag = get_tag();
		if ((Dindex = get_freeR_in_CDTable()) == -1) {
			out.error = MAX_NUM_AGREEMENTS_RUNNING;
			process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
			return;
		}
		CDTable[Dindex].tag = tag;
		memcpy(CDTable[Dindex].elist, in->elist, sizeof(elist_t));
		CDTable[Dindex].tstart = in->tstart;
		CDTable[Dindex].decision = in->decision;
	}
	j = index_eid_in_elist(in->eid, in->elist);
	CDTable[Dindex].vtable[j].is_free = 0;
	memcpy(CDTable[Dindex].vtable[j].value, in->value, sizeof(value_t));

	out.error = 0;
	out.tag = tag;
//temp_time1 = get_vclock();
	process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
	DEBUG_MSG(INTERNAL_LEVEL,"Output from propose sent\n");	
	return;
}

#endif //DIST_AGREEMENT


#ifdef DIST_AGREEMENT
void TTCBdecide(int fifo, decide_in_t *in)
{
	short i, j;
	decide_out_t out;
	TCBtimestamp actTime;
	mask_t bit = 01;
	short all_proposed = 1;
	unsigned long long max_time;
	unsigned short Rindex;
	int MaxIndex; /* This indicates the index of the most voted value */ 
	int MaxValues; /* This indicates the number of values proposed */
	int cycling; /* Just a cycling variable for a loop */
	int ValueFound=0;
	struct {
		value_t value;
		char number; // n. of times this value was proposed 
	} valuestable[MAX_SIZE_ELIST]; // vtable[i] has the value correspondent to the entity in elist[i]
//TCBtimestamp temp_prop, temp_dec, temp_t2;

	out.error = 0;

	Rindex = EIDtoINDEX(in->eid);
	
	if ((i = search_CDTable_tag(in->tag)) == -1) {
		out.error = INVALID_TAG; 	
		DEBUG_MSG(SERVICE_LEVEL," INVALID_TAG\n");
		process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
		return;
	}

	for (j = 0; ((j < MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST)); j++) {
		if (CDTable[i].vtable[j].is_free == 1)
			all_proposed = 0;
	}
	TCBenter_in_critical_section();
	actTime=get_vclock();
	TCBleave_critical_section();
//if ((all_proposed == 1) || (actTime > (CDTable[i].tstart + 13000000))) { 
// NOTA: 13 ms = periodo de envio + recepcao + Tmsg + relogio (Miguel)
/*
  NOTA: Pedro Martins
  
  Miguel, tinhas 13 ms hardcoded, o que nao e nada bom :-) Mudei isso!!!!!
  	
*/ 
	//para ficar coerente como o crash detector
	max_time=READ_PERIOD+1000000+(SEND_PERIOD*(NETWORK_OMISSION_DEGREE+1))+SECURITY_MARGIN_FOR_NODE_CRASH_DETECTION+((2*MAXIMUM_ROUNDTRIP*1000)-(MINIMUM_DELIVERY_DELAY*1000));
	if ((all_proposed == 1) || (actTime > (CDTable[i].tstart + max_time))) { 

	  DEBUG_MSG(DECISAO_LEVEL,"XX ALL_PROPOSED: %d IN_TIME?%d\n", all_proposed, (actTime > CDTable[i].tstart + 50000000));
	  out.mask1 = 0;
	  out.mask2 = 0;

	  switch(CDTable[i].decision) {

	  case TTCB_TBA_RMULTICAST:
	    DEBUG_MSG(DECISAO_LEVEL,"Going for MULTICAST decision\n");
	    if (CDTable[i].vtable[0].is_free == 1) {
	      out.error = DID_NOT_PROPOSE; // first element in elist did not propose
	      DEBUG_MSG(DECISAO_LEVEL," DID_NOT_PROPOSE\n");
	      process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
	      return;
	    }
	    memcpy(out.value, CDTable[i].vtable[0].value, sizeof(value_t));
	    for (j = 0; (j < MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST); j++) {
	      if (CDTable[i].vtable[j].is_free == 0) {
		out.mask2 |= bit;
		if (memcmp(CDTable[i].vtable[j].value, out.value, sizeof(value_t)) == 0)
		  out.mask1 |= bit; 
	      }
	      bit = bit << 1;
	    }
	    DEBUG_MSG(DECISAO_LEVEL,"Chose value [%s]. Sending it.\n", out.value);
	    break;
	    
	  case TTCB_TBA_MAJORITY:
	    DEBUG_MSG(DECISAO_LEVEL,"Going for MAJORITY decision\n");
	    MaxIndex=0;
	    MaxValues=0;
	    ValueFound=0;  
	    DEBUG_MSG(DECISAO_LEVEL,"The valuestable[o].value is %s\n", valuestable[0].value);
	    /* Percorrer a lista de valores e encontrar o mais votado! */
	    for (j = 0; (j < MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST); j++) {

	      ValueFound=0;  
	      DEBUG_MSG(DECISAO_LEVEL,"j is %i\n",j);
	      if (CDTable[i].vtable[j].is_free == 0)
	      {
		if(MaxValues>0)
		{	
		for(cycling=0; cycling < MaxValues;cycling++)
		  {
		    DEBUG_MSG(DECISAO_LEVEL,"Cycling is %i. MaxValues is %i\n", cycling, MaxValues);
		    DEBUG_MSG(DECISAO_LEVEL,"Comparing %s with %s\n", CDTable[i].vtable[j].value, valuestable[cycling].value);
		    /* This is the same value, so just add it to the number of proposed times */ 
		    if ( memcmp(valuestable[cycling].value,CDTable[i].vtable[j].value, sizeof(value_t))==0)
		      {
			DEBUG_MSG(DECISAO_LEVEL,"They're equal!\n");
			/* Increase the number of times this value was proposed */
			valuestable[cycling].number++;
			ValueFound=1;
			/* If this value was proposed more times than the old max, make it the new max */
			if (valuestable[cycling].number > valuestable[MaxIndex].number) MaxIndex=cycling;
			DEBUG_MSG(DECISAO_LEVEL,"MaxIndex is %i\n",MaxIndex);
		      } 
		    
		  }
		}
		
		if(ValueFound == 0) /* This value was not in our table yet, so add it to the table */
		  {
		    DEBUG_MSG(DECISAO_LEVEL,"Value %s was not found. Inserting it.\n",CDTable[i].vtable[j].value);
		    memcpy(valuestable[MaxValues].value, CDTable[i].vtable[j].value, sizeof(value_t));
		    valuestable[MaxValues].number=1;
		    MaxValues++;
		  }   
	      }
	    }
	    
	    memcpy(out.value, valuestable[MaxIndex].value, sizeof(value_t));
	    DEBUG_MSG(DECISAO_LEVEL,"The most voted value was %s\n",out.value);
	    for (j = 0; (j < MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST); j++) {
	      if (CDTable[i].vtable[j].is_free == 0) {
		out.mask2 |= bit;
		if (memcmp(CDTable[i].vtable[j].value, out.value, sizeof(value_t)) == 0)
		  out.mask1 |= bit; 
	      }
	      bit = bit << 1;
	    }
	    
	    break;
	    
	    
	  case TTCB_TBA_AND:
	    DEBUG_MSG(DECISAO_LEVEL,"Going for AND decision\n");
	    
	    // Make the out.value a null element. Since we are talking
	    // about 160 bits, there is no long enough constant, so we
	    // must make it like this.
	    for(cycling=0;cycling<sizeof(value_t); cycling++)
	      {
		out.value[cycling]=0xFF;
	      }

	    for (j = 0; (j < MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST); j++) 
	      {

			if (CDTable[i].vtable[j].is_free == 0)
		  {
		    for(cycling=0;cycling<sizeof(value_t); cycling++)
		      {
			out.value[cycling] &= CDTable[i].vtable[j].value[cycling];
		      }
		  }
	      }
	    
	    DEBUG_MSG(DECISAO_LEVEL,"The most voted value was %s\n",out.value);
	    for (j = 0; (j < MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST); j++) {
	      if (CDTable[i].vtable[j].is_free == 0) {
		out.mask2 |= bit;
		if (memcmp(CDTable[i].vtable[j].value, out.value, sizeof(value_t)) == 0)
		  out.mask1 |= bit; 
	      }
	      bit = bit << 1;
	    }
	    
	    break;

	  case TTCB_TBA_OR:
	    DEBUG_MSG(DECISAO_LEVEL,"Going for AND decision\n");
	    
	    // Make the out.value a null element. Since we are talking
	    // about 160 bits, there is no long enough constant, so we
	    // must make it like this.
	    for(cycling=0;cycling<sizeof(value_t); cycling++)
	      {
		out.value[cycling]=0x00;
	      }

	    for (j = 0; (j < MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST); j++) 
	      {
		if (CDTable[i].vtable[j].is_free == 0)
		  {
		    for(cycling=0;cycling<sizeof(value_t); cycling++)
		      {
			out.value[cycling] |= CDTable[i].vtable[j].value[cycling];
		      }
		  }
	      }
	    
	    DEBUG_MSG(DECISAO_LEVEL,"The most voted value was %s\n",out.value);
	    for (j = 0; (j < MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST); j++) {
	      if (CDTable[i].vtable[j].is_free == 0) {
		out.mask2 |= bit;
		if (memcmp(CDTable[i].vtable[j].value, out.value, sizeof(value_t)) == 0)
		  out.mask1 |= bit; 
	      }
	      bit = bit << 1;
	    }
	    
	    break;

	  case TTCB_TBA_XOR:
	    DEBUG_MSG(DECISAO_LEVEL,"Going for XOR decision\n");
	    
	    ValueFound = 0;
	    // Make the out.value a null element. Since we are talking
	    // about 160 bits, there is no long enough constant, so we
	    // must make it like this.

	    //For the XOR case, we can not have a null element, so this is not necessary.
	    //Instead, we must equal it to the first decided value, and go on from there.

	    //	    for(cycling=0;cycling<sizeof(value_t); cycling++)
	    //{
	    //out.value[cycling]=0xFF;
	    //}

	    for (j = 0; (j < MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST); j++) 
	      {
		if (CDTable[i].vtable[j].is_free == 0)
		  {
		    if(ValueFound==0)
		      {
			memcpy(out.value, CDTable[i].vtable[j].value, sizeof(value_t));
			ValueFound=1;
		      }
		    else
		      {
			for(cycling=0;cycling<sizeof(value_t); cycling++)
			  { 
			    out.value[cycling] ^= CDTable[i].vtable[j].value[cycling];
			  }
		      }
		  }
	      }
	    
	    DEBUG_MSG(DECISAO_LEVEL,"The most voted value was %s\n",out.value);
	    for (j = 0; (j < MAX_SIZE_ELIST) && (CDTable[i].elist[j] != END_ELIST); j++) {
	      if (CDTable[i].vtable[j].is_free == 0) {
		out.mask2 |= bit;
		if (memcmp(CDTable[i].vtable[j].value, out.value, sizeof(value_t)) == 0)
		  out.mask1 |= bit; 
	      }
	      bit = bit << 1;
	    }
	    
	    break;

	  default:
	    DEBUG_MSG(DECISAO_LEVEL,"INVALID DECISION type\n ");
	  }
	  //temp_time3 = get_vclock();
	  //temp_prop = temp_time1 - temp_time0; // Tprop
	  //temp_dec = temp_time3 - temp_time2; // Tdec
	  //temp_t2 = temp_time3 - temp_time0;   // T2
	  //memcpy(out.value, &temp_t2, sizeof(temp_time0));
	  //memcpy(&(out.value[8]), &temp_prop, sizeof(temp_time0));
	  //memcpy(&(out.value[16]), &temp_dec, sizeof(temp_time0));
	  
	  process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
	  
	} else {
	  out.error = AGREEMENT_RUNNING;
	  DEBUG_MSG(SERVICE_LEVEL,"AGREEMENT_RUNNING\n");
	  process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
	}

	DEBUG_MSG(INTERNAL_LEVEL,"Going to return from TTCBdecide.\n");
	return;
}
#endif // DIST_AGREEMENT

/*
+-----------------------+
| random number service |
+-----------------------+
*/

#ifdef LOCAL_RANDOM
void TTCBrandom(int fifo, eid_in_t *in)
{
	rand_out_t out;
	short Rindex;

	out.error = 0;
	Rindex = EIDtoINDEX(in->eid);
	get_random_bytes(out.random, 16);
	process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
}
#endif


/*
+----------------------------+
| local TTCB crash detection |
+----------------------------+
*/

//VER ISTO!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

void TTCBcrashedLocalTTCBid(int fifo, localTTCBid_in_t *in)
{
	error_out_t out;
	short Rindex;

	Rindex = EIDtoINDEX(in->eid);
	out.error = system_view[in->localTTCBid];
	process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
}


void TTCBcrashedLocalTTCBeid(int fifo, eid_in_t *in)
{
	error_out_t out;
	localTTCBid_t ltid;
	short Rindex;

	Rindex = EIDtoINDEX(in->eid);
	ltid = EIDtoLTID(in->eid);  // verificar se eid esta nos limites
	out.error = system_view[ltid];
	process_output(fifo, (out_hdr_t *)&out, sizeof(out), in->eid, entities_data[Rindex].protection, entities_data[Rindex].secret, entities_data[Rindex].seq);
}


#endif  //TTCB
