
/*
 *	RTloopback.c
 *	based on <drivers/net/loopback.c>	
 *
 *	RT-INET 1.0
 *
 *	(C)2002, Navigators Group	<www.navigators.di.fc.ul.pt>
 *		 Pedro Martins		<pmartins@di.fc.ul.pt>
 *	Loopback driver for RTIP	
 *
 * 	Last Review Date: XX
*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/socket.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/in.h>
#include <linux/init.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/if_ether.h>	/* For the statistics structure. */
#include <linux/if_arp.h>	/* For ARPHRD_ETHER */

#include "rtip/RTdev.h"
#include "rtip/RTetherdevice.h"

static int RTloopback_xmit(struct RTsk_buff *skb, struct RTnet_device *dev)
{
	struct RTnet_device_stats *stats = (struct RTnet_device_stats *)dev->priv;

#if 0
	if(atomic_read(&skb->users) != 1)
	{
	  	struct sk_buff *skb2=skb;
	  	skb=skb_clone(skb, GFP_ATOMIC);		/* Clone the buffer */
	  	if(skb==NULL) {
			kfree_skb(skb2);
			return 0;
		}
	  	kfree_skb(skb2);
	}
	else
		skb_orphan(skb);
	
	skb->protocol=eth_type_trans(skb,dev);
	skb->dev=dev;
#endif
	skb->protocol=RTeth_type_trans(skb,dev);
	skb->ip_summed = CHECKSUM_UNNECESSARY;
	dev->last_rx = jiffies;
	stats->rx_bytes+=skb->len;
	stats->tx_bytes+=skb->len;
	stats->rx_packets++;
	stats->tx_packets++;
	skb->dev=dev;
	RTnetif_rx(skb);

	return(0);
}

static struct RTnet_device_stats *RTget_stats(struct RTnet_device *dev)
{
	return (struct RTnet_device_stats *)dev->priv;
}

/* Initialize the rest of the LOOPBACK device. */
int RTloopback_init(struct RTnet_device *dev)
{
	dev->mtu		= 1500;   
	dev->hard_start_xmit	= RTloopback_xmit;
	dev->hard_header	= RTeth_header;
	dev->hard_header_len	= ETH_HLEN;		/* 14			*/
	dev->addr_len		= ETH_ALEN;		/* 6			*/
	dev->tx_queue_len	= 0;
	dev->type		= ARPHRD_LOOPBACK;	/* 0x0001		*/
	dev->flags		= IFF_LOOPBACK;
	dev->features		= NETIF_F_SG|NETIF_F_FRAGLIST|NETIF_F_NO_CSUM|NETIF_F_HIGHDMA;
	atomic_set(&(dev->is_receiving),0); 
	dev->priv = kmalloc(sizeof(struct RTnet_device_stats), GFP_KERNEL);
	if (dev->priv == NULL)
			return -ENOMEM;
	memset(dev->priv, 0, sizeof(struct RTnet_device_stats));
	dev->get_stats = RTget_stats;

	/*
	 *	Fill in the generic fields of the device structure. 
	 */
   
	return(0);
};
