/*
 *	RTnet_init.c
 *	based on <drivers/net/net_init.c>	
 *
 *	RT-INET 1.0
 *
 *	(C)2002, Navigators Group	<www.navigators.di.fc.ul.pt>
 *		 Pedro Martins		<pmartins@di.fc.ul.pt>
 *	Device initialization routines
 *
 * 	Last Review Date: XX
*/

#include "rtip/RTnet_init.h"
#include "rtip/RTetherdevice.h"
#include "rtip/RTnetdevice.h"

static struct RTnet_device *RTalloc_netdev(int sizeof_priv, const char *mask,
				       void (*setup)(struct RTnet_device *))
{
	struct RTnet_device *dev;
	int alloc_size;

	/* ensure 32-byte alignment of the private area */
	alloc_size = sizeof (*dev) + sizeof_priv + 31;

	dev = (struct RTnet_device *) kmalloc (alloc_size, GFP_KERNEL);
	if (dev == NULL)
	{
		printk(KERN_ERR "alloc_dev: Unable to allocate device memory.\n");
		return NULL;
	}

	memset(dev, 0, alloc_size);

	if (sizeof_priv)
		dev->priv = (void *) (((long)(dev + 1) + 31) & ~31);

	setup(dev);
	strcpy(dev->name, mask);

	return dev;
}



/**
 * RTalloc_etherdev - Allocates and sets up an ethernet device
 * @sizeof_priv: Size of additional driver-private structure to be allocated
 *	for this ethernet device
 *
 * Fill in the fields of the device structure with ethernet-generic
 * values. Basically does everything except registering the device.
 *
 * Constructs a new net device, complete with a private data area of
 * size @sizeof_priv.  A 32-byte (not bit) alignment is enforced for
 * this private data area.
 */

struct RTnet_device *RTalloc_etherdev(int sizeof_priv)
{
	return RTalloc_netdev(sizeof_priv, "rteth%d", RTether_setup);
}

//EXPORT_SYMBOL(RTalloc_etherdev);

#if 0
static int eth_mac_addr(struct net_device *dev, void *p)
{
	struct sockaddr *addr=p;
	if (netif_running(dev))
		return -EBUSY;
	memcpy(dev->dev_addr, addr->sa_data,dev->addr_len);
	return 0;
}

static int eth_change_mtu(struct net_device *dev, int new_mtu)
{
	if ((new_mtu < 68) || (new_mtu > 1500))
		return -EINVAL;
	dev->mtu = new_mtu;
	return 0;
}

#endif

void RTether_setup(struct RTnet_device *dev)
{
	/* Fill in the fields of the device structure with ethernet-generic values.
	   This should be in a common file instead of per-driver.  */
	
#if 0
	dev->change_mtu		= eth_change_mtu;
#endif
	dev->hard_header	= RTeth_header;
#if 0
	dev->rebuild_header 	= eth_rebuild_header;
	dev->set_mac_address 	= eth_mac_addr;
	dev->hard_header_cache	= eth_header_cache;
	dev->header_cache_update= eth_header_cache_update;
	dev->hard_header_parse	= eth_header_parse;
#endif

	dev->type		= ARPHRD_ETHER;
	dev->hard_header_len 	= ETH_HLEN;
	dev->mtu		= 1500; /* eth_mtu */
	dev->addr_len		= ETH_ALEN;
	atomic_set(&(dev->is_receiving),0); 
#if 0
	dev->tx_queue_len	= 100;	/* Ethernet wants good queues */	
#endif
	
	memset(dev->broadcast,0xFF, ETH_ALEN);

	/* New-style flags. */
	dev->flags		= IFF_BROADCAST|IFF_MULTICAST;
}

//EXPORT_SYMBOL(RTether_setup);


int RTregister_netdev(struct RTnet_device *dev)
{
	int err;

	//rtnl_lock();

	err = RTdev_alloc_name(dev, dev->name);
	if (err < 0){
		goto out;
	}
	
	err = RTregister_netdevice(dev);

out:
	//rtnl_unlock();
	return err;
}

void RTunregister_netdev(struct RTnet_device *dev)
{
	//rtnl_lock();
	RTunregister_netdevice(dev);
	//rtnl_unlock();
}

//EXPORT_SYMBOL(RTregister_netdev);
//EXPORT_SYMBOL(RTunregister_netdev);

