
/*
 *	RTskbuff.c
 *	based on <net/core/skbuff.c>	
 *
 *	RT-INET 1.0
 *
 *	(C)2002, Navigators Group	<www.navigators.di.fc.ul.pt>
 *		 Pedro Martins		<pmartins@di.fc.ul.pt>
 *	Real-Time Socket Buffers	
 *
 * 	Last Review Date: XX
*/

#include "rtip/RTskbuff.h"

/**
 *	RTalloc_skb	-	allocate a network buffer
 *	@size: size to allocate
 *
 *	Allocate a new &RTsk_buff. The returned buffer has no headroom and a
 *	tail room of size bytes. The object has a reference count of one.
 *	The return is the buffer. On a failure the return is %NULL.
 *
 */
 
object_list RTskbuffs; 
 
struct RTsk_buff *RTalloc_skb(void){

	RTsk_buff *skb;
	pt_obj_list pt;	

	skb = NULL;
	pt = get_obj_list_item(&RTskbuffs);
	if (pt!=NULL)
		skb = (RTsk_buff*)pt->object;
	if (skb == NULL) {
			return NULL;
	}
	skb->truesize = sizeof(struct RTsk_buff);

	/* Load the data pointers. */
	skb->data = skb->bdata;
	skb->head = skb->bdata;
	skb->tail =  skb->bdata;
	skb->end = skb->bdata + SKBUFF_PKT_SIZE;

	/* Set up other state */
	skb->len = 0;
	skb->data_len = 0;
	skb->obj_list = pt;

	//skb->stamp.tv_sec=0;	/* No idea about time */
	skb->pkt_type = PACKET_HOST;	/* Default type */
	skb->ip_summed = 0;
//	skb->clone_skb = 0;
	skb->priority = 0;
//	skb->cloned = 0;
//	atomic_set(&skb->users, 1);

	return skb;
};


/**
 *	RTskb_copy	-	duplicate an RTsk_buff
 *	@skb: buffer to clone
 *
 */

struct RTsk_buff *RTskb_copy(struct RTsk_buff *skb)
{
	struct RTsk_buff *n;
	pt_obj_list pt;	

	n = NULL;
	pt = get_obj_list_item(&RTskbuffs);
	if (pt!=NULL){
		n = (RTsk_buff*)pt->object;
		memcpy(n,skb,sizeof(struct RTsk_buff));
		n->data = n->bdata + (skb->data - skb->bdata);
		n->head = n->bdata + (skb->head - skb->bdata);;
		n->tail = n->bdata + (skb->tail - skb->bdata);
		n->end =  n->bdata + (skb->end - skb->bdata);
		n->obj_list = pt;
	}		
	return n;
}

int initRTskbuffs(int n_buffs){
	if (init_obj_list(&RTskbuffs,n_buffs,sizeof(RTsk_buff))){
		return -ENOMEM;
	}
	return 0;
}

void endRTskbuffs(void){
	free_obj_list(&RTskbuffs);
}



