/*	
	Display (T)TCB parameters

	Pedro Martins <pmartins@di.fc.ul.pt>

--[TTCB 1.1]----------------------------------------------------

    (Trusted) Timely Computing Base 1.1
    (C) 1999-2003, Navigators Group 
    		   http://www.navigators.di.fc.ul.pt
    
-----------------------------------------------------------------	
*/

#include "tcb/system_defs.h"
#include "tcb/driver_defs.h"
#include "tcb/messages_formats.h"
#ifdef TTCB
#include "tcb/ttcb_module.h"
#endif

int main()
{
	int max_mtu;
	int net_mtu=MTU,max_fields,max_items,max_fields_agreement;
	
	max_mtu=NET_FIFO_SIZE/NUMBER_OF_MACHINES;
	
	if (max_mtu>net_mtu)
	 max_mtu=net_mtu;
	
//	max_mtu=max_mtu/(NETWORK_OMISSION_DEGREE+1);	

	printf("\n------------------\n");
	printf("TCB specifications\n");
	printf("------------------\n");
#ifdef DISTRIBUTED
	printf(".Control channel device : <%s>\n",NET_DEVICE_NAME);
	printf(".Control channel MTU : <%u> bytes.\n",MTU);
	printf(".Control channel network card fifo size : <%u> bytes.\n",NET_FIFO_SIZE);
	printf(".Transmission period : <%u> us.\n",SEND_PERIOD/1000);
	printf(".Omission degree of the control channel : <%u>\n",NETWORK_OMISSION_DEGREE);
	printf(".Send period : <%u> us.\n",(SEND_PERIOD*(NETWORK_OMISSION_DEGREE+1))/1000);
	printf(".Read period : <%u> us.\n",READ_PERIOD/1000);
	printf(".Maximum packet size : <%u>\n",max_mtu);
	printf(".Minimum delivery delay on the control channel : <%u> us.\n",MINIMUM_DELIVERY_DELAY);
	printf(".Maximum round-trip value allowed for the control channel : <%u> us.\n",MAXIMUM_ROUNDTRIP);
	#ifdef CONFIG_SELF_CHECKING_CONTROL_CHANNEL_SIMULATED	
	printf(".Self-Ckecking mode for the control channel : <simulated>\n");
#elif CONFIG_SELF_CHECKING_CONTROL_CHANNEL_REAL
	printf(".Self-Ckecking mode for the control channel : <real>\n");
#else
	printf(".Self-Ckecking mode for the control channel : <none>\n");	
#endif	
#else
	printf(".The TTCB is not configured for a distributed operation\n");
#endif //DISTRIBUTED	
	printf(".Number of machines in the system: <%u>\n",NUMBER_OF_MACHINES);

#if defined(DIST_AGREEMENT) && defined(DIST_FAULT)
	max_mtu = max_mtu / 2;  //half to header+FD, another half to timestamp+Agreement
	max_fields_agreement = (max_mtu / sizeof(CSTable_field));
#endif

#if defined(DIST_AGREEMENT) && !defined(DIST_FAULT)
	max_fields_agreement = ((max_mtu-sizeof(msgtr_hdr)) / sizeof(CSTable_field));
#endif

#ifdef DIST_FAULT
	max_fields=(max_mtu-sizeof(msgtr_hdr))/sizeof(info_field);
	max_items=max_fields/(NUMBER_OF_MACHINES+1);		
#endif

#ifdef DIST_FAULT
	printf(".Maximum number of distributed failure detection protocol info fields allowed to be relayed in one send round : <%u>\n",max_fields);
	printf(".Maximum number of simultaneous distributed failure detection services allowed   by the TCB : <%u>\n",max_items);
	printf(".Maximum number of distributed failure detection services allowed by the TCB per send period : <%u>\n",max_items);
#else
	printf(".The distributed timing failure detection service is not configured on this TTCB\n");
#endif
#ifdef LOCAL_FAULT
	printf(".Maximum number of simultaneous local failure detections allowed by the TCB: <%u>\n",MAX_LOCAL_FD);
#else
	printf(".The local timing failure detection service is not configured on this TTCB\n");
#endif
#ifdef LOCAL_TIMELY 
	printf(".Maximum number of simultaneous timely executions allowed by the TCB: <%u>\n",MAX_EXEC);	
#else
	printf(".The timely execution service is not configured on this TTCB\n");
#endif
#ifdef TTCB
	printf("----------\n");
	printf("TTCB specs\n");
	printf("----------\n");	
	
#endif
	printf("--------------------------------------------------------------------------------\n");	
	return 0;
}
